import json
import logging
import os
from typing import List, Tuple, Type

import pandas as pd
from scipy import sparse as sps

from irspack.dataset.movielens import MovieLens1MDataManager
from irspack.evaluator import EvaluatorWithColdUser
from irspack.optimizers import (  # SLIMOptimizer,; MultVAEOptimizer,
    AsymmetricCosineKNNOptimizer,
    BaseOptimizer,
    CosineKNNOptimizer,
    DenseSLIMOptimizer,
    IALSOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    TopPopOptimizer,
    TverskyIndexKNNOptimizer,
)
from irspack.split import split_dataframe_partial_user_holdout

os.environ["OMP_NUM_THREADS"] = "8"
os.environ["IRSPACK_NUM_THREADS_DEFAULT"] = "8"

if __name__ == "__main__":

    BASE_CUTOFF = 20

    data_manager = MovieLens1MDataManager()
    df_all = data_manager.read_interaction()

    data_all, _ = split_dataframe_partial_user_holdout(
        df_all,
        "userId",
        "movieId",
        test_user_ratio=0.2,
        val_user_ratio=0.2,
        heldout_ratio_test=0.5,
        heldout_ratio_val=0.5,
    )

    data_train = data_all["train"]
    data_val = data_all["val"]
    data_test = data_all["test"]

    X_train_val_all: sps.csr_matrix = sps.vstack(
        [data_train.X_all, data_val.X_all], format="csr"
    )
    valid_evaluator = EvaluatorWithColdUser(
        input_interaction=data_val.X_train,
        ground_truth=data_val.X_test,
        cutoff=BASE_CUTOFF,
    )
    test_evaluator = EvaluatorWithColdUser(
        input_interaction=data_test.X_train,
        ground_truth=data_test.X_test,
        cutoff=BASE_CUTOFF,
    )

    test_results = []
    validation_results = []

    test_configs: List[Tuple[Type[BaseOptimizer], int]] = [
        (TopPopOptimizer, 1),
        (CosineKNNOptimizer, 40),
        (AsymmetricCosineKNNOptimizer, 40),
        (TverskyIndexKNNOptimizer, 40),
        (P3alphaOptimizer, 40),
        (RP3betaOptimizer, 40),
        (IALSOptimizer, 40),
        (DenseSLIMOptimizer, 20),
        # (SLIMOptimizer, 40), time consuming
        # (MultVAEOptimizer, 5),
    ]
    for optimizer_class, n_trials in test_configs:
        recommender_name = optimizer_class.recommender_class.__name__
        optimizer: BaseOptimizer = optimizer_class(
            data_train.X_all,
            valid_evaluator,
            metric="ndcg",
        )
        (best_param, validation_result_df) = optimizer.optimize(
            timeout=14400, n_trials=n_trials
        )
        validation_result_df["recommender_name"] = recommender_name
        validation_results.append(validation_result_df)
        pd.concat(validation_results).to_csv(f"validation_scores.csv")
        test_recommender = optimizer.recommender_class(X_train_val_all, **best_param)
        test_recommender.learn()
        test_scores = test_evaluator.get_scores(test_recommender, [5, 10, 20])

        test_results.append(
            dict(name=recommender_name, best_param=best_param, **test_scores)
        )
        with open("test_results.json", "w") as ofs:
            json.dump(test_results, ofs, indent=2)



import warnings
from collections import OrderedDict
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np

from irspack._evaluator import EvaluatorCore, Metrics
from irspack.definitions import DenseScoreArray, InteractionMatrix
from irspack.utils import get_n_threads

if TYPE_CHECKING:
    from .recommenders import base as base_recommender


class TargetMetric(Enum):
    NDCG = "ndcg"
    RECALL = "recall"
    HIT = "hit"


METRIC_NAMES = [
    "hit",
    "recall",
    "ndcg",
    "map",
    "gini_index",
    "entropy",
    "appeared_item",
]


class Evaluator:
    """Evaluates recommenders' performance against validation set.

    Args:
        ground_truth (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]): The held-out ground-truth.
        offset (int): Where the validation target user block begins.
            Often the validation set is defined for a subset of users.
            When offset is not 0, we assume that the users with validation
            ground truth corresponds to X_train[offset:] where X_train
            is the matrix feeded into the recommender class.
        cutoff (int, optional): Controls the number of recommendation.
            Defaults to 10.
        target_metric (str, optional): Optimization target metric.
            Defaults to "ndcg".
        recommendable_items (Optional[List[int]], optional): Global recommendable items. Defaults to None.
            If this parameter is not None, evaluator will be concentrating on
            the recommender's score output for these recommendable_items,
            and compute the ranking performance within this subset.
        per_user_recommendable_items (Optional[List[List[int]]], optional):
            Similar to `recommendable_items`, but this time the recommendable items can vary among users. Defaults to None.
        n_threads (int, optional): Number of threads to sort the score and compute the
            evaluation metrics. Defaults to 1.
        mb_size (int, optional): The rows of chunked user score. Defaults to 1024.
    """

    n_users: int
    n_items: int

    def __init__(
        self,
        ground_truth: InteractionMatrix,
        offset: int = 0,
        cutoff: int = 10,
        target_metric: str = "ndcg",
        recommendable_items: Optional[List[int]] = None,
        per_user_recommendable_items: Optional[List[List[int]]] = None,
        n_threads: Optional[int] = None,
        mb_size: int = 1024,
    ) -> None:

        ground_truth = ground_truth.tocsr().astype(np.float64)
        ground_truth.sort_indices()
        if recommendable_items is None:
            if per_user_recommendable_items is None:
                recommendable_items_arg: List[List[int]] = []
            else:
                recommendable_items_arg = per_user_recommendable_items
        else:
            recommendable_items_arg = [recommendable_items]

        self.core = EvaluatorCore(ground_truth, recommendable_items_arg)
        self.offset = offset
        self.n_users = ground_truth.shape[0]
        self.n_items = ground_truth.shape[1]
        self.target_metric = TargetMetric(target_metric)
        self.cutoff = cutoff
        self.n_threads = get_n_threads(n_threads)
        self.mb_size = mb_size

    def get_score(self, model: "base_recommender.BaseRecommender") -> Dict[str, float]:
        """Compute the score with the cutoff being `self.cutoff`.

        Args:
            model : The evaluated recommender.

        Returns:
            metric values.
        """
        return self._get_scores_as_list(model, [self.cutoff])[0]

    def get_scores(
        self, model: "base_recommender.BaseRecommender", cutoffs: List[int]
    ) -> Dict[str, float]:
        """Compute the score with the specified cutoffs.

        Args:
            model : The evaluated recommender.
            cutoffs : for each value in cutoff, the class computes
                the metric values.

        Returns:
            The Resulting metric values. This time, the result
            will look like ``{"ndcg@20": 0.35, "map@20": 0.2, ...}``.
        """

        result: Dict[str, float] = OrderedDict()
        scores = self._get_scores_as_list(model, cutoffs)
        for cutoff, score in zip(cutoffs, scores):
            for metric_name in METRIC_NAMES:
                result[f"{metric_name}@{cutoff}"] = score[metric_name]
        return result

    def _get_scores_as_list(
        self, model: "base_recommender.BaseRecommender", cutoffs: List[int]
    ) -> List[Dict[str, float]]:
        if self.offset + self.n_users > model.n_users:
            raise ValueError("evaluator offset + n_users exceeds the model's n_users.")
        if self.n_items != model.n_items:
            raise ValueError("The model and evaluator assume different n_items.")
        n_items = self.n_items
        metrics: List[Metrics] = []
        for c in cutoffs:
            metrics.append(Metrics(n_items))

        block_start = self.offset
        n_validated = self.n_users
        block_end = block_start + n_validated
        mb_size = self.mb_size

        for chunk_start in range(block_start, block_end, mb_size):
            chunk_end = min(chunk_start + mb_size, block_end)
            try:
                # try faster method
                scores = model.get_score_remove_seen_block(chunk_start, chunk_end)
            except NotImplementedError:
                # block-by-block
                scores = model.get_score_remove_seen(np.arange(chunk_start, chunk_end))
            for i, c in enumerate(cutoffs):
                chunked_metric = self.core.get_metrics(
                    scores, c, chunk_start - self.offset, self.n_threads, False
                )
                metrics[i].merge(chunked_metric)

        return [item.as_dict() for item in metrics]


class EvaluatorWithColdUser(Evaluator):
    """Evaluates recommenders' performance against cold (unseen) users.

    Args:
        input_interaction (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]): The cold-users' known interaction
            with the items.
        ground_truth (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]): The held-out ground-truth.
        offset (int): Where the validation target user block begins.
            Often the validation set is defined for a subset of users.
            When offset is not 0, we assume that the users with validation
            ground truth corresponds to X_train[offset:] where X_train
            is the matrix feeded into the recommender class.
        cutoff (int, optional): Controls the number of recommendation.
            Defaults to 10.
        target_metric (str, optional): Optimization target metric.
            Defaults to "ndcg".
        recommendable_items (Optional[List[int]], optional): Global recommendable items. Defaults to None.
            If this parameter is not None, evaluator will be concentrating on
            the recommender's score output for these recommendable_items,
            and compute the ranking performance within this subset.
        per_user_recommendable_items (Optional[List[List[int]]], optional):
            Similar to `recommendable_items`, but this time the recommendable items can vary among users. Defaults to None.
        n_threads (int, optional): Number of threads to sort the score and compute the
            evaluation metrics. Defaults to 1.
        mb_size (int, optional): The rows of chunked user score. Defaults to 1024.
    """

    def __init__(
        self,
        input_interaction: InteractionMatrix,
        ground_truth: InteractionMatrix,
        cutoff: int = 10,
        target_metric: str = "ndcg",
        recommendable_items: Optional[List[int]] = None,
        per_item_recommendable_items: Optional[List[List[int]]] = None,
        n_threads: Optional[int] = None,
        mb_size: int = 1024,
    ):

        super(EvaluatorWithColdUser, self).__init__(
            ground_truth,
            0,
            cutoff,
            target_metric,
            recommendable_items,
            per_item_recommendable_items,
            n_threads,
            mb_size,
        )
        self.input_interaction = input_interaction

    def _get_scores_as_list(
        self,
        model: "base_recommender.BaseRecommender",
        cutoffs: List[int],
    ) -> List[Dict[str, float]]:

        n_items = model.n_items
        metrics: List[Metrics] = []
        for c in cutoffs:
            metrics.append(Metrics(n_items))

        block_start = self.offset
        n_validated = self.n_users
        block_end = block_start + n_validated
        mb_size = self.mb_size

        for chunk_start in range(block_start, block_end, mb_size):
            chunk_end = min(chunk_start + mb_size, block_end)
            scores = model.get_score_cold_user_remove_seen(
                self.input_interaction[chunk_start:chunk_end]
            )
            if not scores.flags.c_contiguous:
                warnings.warn(
                    "Found col-major(fortran-style) score values.\n"
                    "Transforming it to row-major score matrix."
                )
                scores = np.ascontiguousarray(scores, dtype=np.float64)

            for i, c in enumerate(cutoffs):
                chunked_metric = self.core.get_metrics(
                    scores, c, chunk_start, self.n_threads, False
                )
                metrics[i].merge(chunked_metric)

        return [item.as_dict() for item in metrics]



import os

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.recommenders.knn import (
    AsymmetricCosineKNNRecommender,
    CosineKNNRecommender,
    JaccardKNNRecommender,
    TverskyIndexKNNRecommender,
)
from irspack.recommenders.p3 import P3alphaRecommender
from irspack.recommenders.rp3 import RP3betaRecommender

X_small = sps.csr_matrix(
    np.asfarray([[1, 1, 2, 3, 4], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]])
)
X_many = np.random.rand(888, 512)
X_many[X_many <= 0.9] = 0
X_many[X_many > 0.9] = 1
X_many = sps.csr_matrix(X_many)
X_many.sort_indices()

X_many_dense = sps.csr_matrix(np.random.rand(133, 245))


@pytest.mark.parametrize(
    "X, normalize", [(X_many, True), (X_small, False), (X_many_dense, True)]
)
def test_cosine(X: sps.csr_matrix, normalize: bool) -> None:
    rec = CosineKNNRecommender(
        X, shrinkage=0, n_threads=5, top_k=X.shape[1], normalize=normalize
    )
    rec.learn()
    sim = rec.W.toarray()
    manual = X.T.toarray()  # I x U
    norm = (manual ** 2).sum(axis=1) ** 0.5
    manual = manual.dot(manual.T)
    if normalize:
        denom = norm[:, None] * norm[None, :] + 1e-6
        manual /= denom
    np.fill_diagonal(manual, 0)
    np.testing.assert_allclose(
        sim,
        manual,
    )


@pytest.mark.parametrize("X", [X_many, X_small, X_many_dense])
def test_jaccard(X: sps.csr_matrix) -> None:
    rec = JaccardKNNRecommender(X, shrinkage=0, top_k=X.shape[1], n_threads=1)
    rec.learn()
    sim = rec.W.toarray()
    X_bin = X.copy()
    X_bin.sort_indices()
    X_bin.data[:] = 1
    manual = X_bin.T.toarray()  # I x U
    norm = manual.sum(axis=1)
    manual = manual.dot(manual.T)
    denom = norm[:, None] + norm[None, :] - manual + 1e-6
    denom[denom <= 1e-10] = 1e-10
    manual = manual / denom
    np.fill_diagonal(manual, 0)
    np.testing.assert_allclose(sim, manual)


@pytest.mark.parametrize(
    "X, alpha, shrinkage",
    [(X_many, 0.5, 0.0), (X_small, 0.7, 1.0), (X_many_dense, 0.01, 3)],
)
def test_asymmetric_cosine(X: sps.csr_matrix, alpha: float, shrinkage: float) -> None:
    rec = AsymmetricCosineKNNRecommender(
        X, shrinkage=shrinkage, alpha=alpha, n_threads=1, top_k=X.shape[1]
    )
    rec.learn()
    sim = rec.W.toarray()

    manual = X.T.toarray()  # I x U
    norm = (manual ** 2).sum(axis=1)
    norm_alpha = np.power(norm, alpha)
    norm_1malpha = np.power(norm, 1 - alpha)
    manual_sim = manual.dot(manual.T)
    denom = norm_alpha[:, None] * norm_1malpha[None, :] + 1e-6 + shrinkage
    manual_sim /= denom
    np.fill_diagonal(manual_sim, 0)
    np.testing.assert_allclose(
        sim,
        manual_sim,
    )


@pytest.mark.parametrize(
    "X, alpha, beta, shrinkage",
    [(X_small, 0.7, 2, 0.0), (X_many, 0.5, 0.5, 0), (X_many_dense, 0.01, 0, 3)],
)
def test_tversky_index(
    X: sps.csr_matrix, alpha: float, beta: float, shrinkage: float
) -> None:
    RNS = np.random.RandomState(0)
    rec = TverskyIndexKNNRecommender(
        X, shrinkage=shrinkage, alpha=alpha, beta=beta, n_threads=1, top_k=X.shape[1]
    )
    rec.learn()
    sim = rec.W.toarray()
    tested_index_row = RNS.randint(0, sim.shape[0], size=100)
    tested_index_col = RNS.randint(0, sim.shape[0], size=100)
    X_csc = X.tocsc()
    X_csc.sorted_indices()
    for i, j in zip(tested_index_row, tested_index_col):
        if i == j:
            continue
        computed = sim[i, j]
        U_i = set(X_csc[:, i].nonzero()[0])
        U_j = set(X_csc[:, j].nonzero()[0])
        intersect = U_i.intersection(U_j)
        Ui_minus_Uj = U_i.difference(U_j)
        Uj_minus_Ui = U_j.difference(U_i)
        target = len(intersect) / (
            len(intersect)
            + alpha * len(Ui_minus_Uj)
            + beta * len(Uj_minus_Ui)
            + shrinkage
            + 1e-6
        )
        assert computed == pytest.approx(target)


@pytest.mark.parametrize("X", [X_many, X_small])
def test_topk(X: sps.csr_matrix) -> None:
    rec = CosineKNNRecommender(X, shrinkage=0, top_k=30, n_threads=5)
    rec.learn()
    sim = rec.W.toarray()
    assert np.all((sim > 0).sum(axis=1) <= 30)


@pytest.mark.parametrize("X, alpha", [(X_small, 0.001), (X_many_dense, 2), (X_many, 1)])
def test_p3(X: sps.csr_matrix, alpha: float) -> None:
    rec = P3alphaRecommender(X, alpha=alpha, n_threads=4)
    rec.learn()
    W = rec.W.toarray()

    P_ui = np.power(X.toarray(), alpha)
    P_iu = np.power(X.T.toarray(), alpha)

    def zero_or_1(X: np.ndarray) -> np.ndarray:
        X = X.copy()
        X[X == 0] = 1
        return X

    P_ui /= zero_or_1(P_ui.sum(axis=1))[:, None]
    P_iu /= zero_or_1(P_iu.sum(axis=1))[:, None]
    W_man = P_iu.dot(P_ui)
    np.testing.assert_allclose(W, W_man)

    rec_norm = P3alphaRecommender(
        X, alpha=alpha, n_threads=4, top_k=2, normalize_weight=True
    )
    rec_norm.learn()
    # rec
    W_sum = rec_norm.W.sum(axis=1).A1
    for w in W_sum:
        assert w == pytest.approx(1.0)


@pytest.mark.parametrize(
    "X, alpha, beta", [(X_small, 0.001, 3), (X_many_dense, 2, 5), (X_many, 1, 0.2)]
)
def test_rp3(X: sps.csr_matrix, alpha: float, beta: float) -> None:
    rec = RP3betaRecommender(X, alpha=alpha, beta=beta, n_threads=4)
    rec.learn()
    W = rec.W.toarray()
    W_sum = W.sum(axis=1)
    W_sum = W_sum[W_sum >= 0]

    popularity = X.sum(axis=0).A1.ravel() ** beta

    def zero_or_1(X: np.ndarray) -> np.ndarray:
        X = X.copy()
        X[X == 0] = 1
        return X

    P_ui = np.power(X.toarray(), alpha)
    P_iu = np.power(X.T.toarray(), alpha)

    P_ui /= zero_or_1(P_ui.sum(axis=1))[:, None]
    P_iu /= zero_or_1(P_iu.sum(axis=1))[:, None]
    W_man = P_iu.dot(P_ui)

    # p_{ui} ^{RP3} = p_{ui} ^{P3} / popularity_i ^ beta
    from sklearn.preprocessing import normalize

    W_man = W_man / zero_or_1(popularity)[None, :]
    # W_man = normalize(W_man, axis=1, norm="l1")
    # print(W_man.sum(axis=1))
    np.testing.assert_allclose(W, W_man)

    rec_norm = RP3betaRecommender(
        X, alpha=alpha, n_threads=4, top_k=2, normalize_weight=True
    )
    rec_norm.learn()
    # rec
    W_sum = rec_norm.W.sum(axis=1).A1
    for w in W_sum:
        assert w == pytest.approx(1.0)


def test_raise_shrinkage() -> None:
    with pytest.raises(ValueError):
        _ = P3alphaRecommender(X_many.T, alpha=1.0, n_threads=0)
        _.learn()

    with pytest.raises(ValueError):
        _ = P3alphaRecommender(X_many.T, alpha=-1.0, n_threads=1)
        _.learn()

    with pytest.raises(ValueError):
        rec = AsymmetricCosineKNNRecommender(X_many, 0, alpha=1.5)
        rec.learn()


@pytest.mark.parametrize(
    "X",
    [X_small],
)
def test_n_threads(X: sps.csr_matrix) -> None:
    os.environ["IRSPACK_NUM_THREADS_DEFAULT"] = "31"
    rec = CosineKNNRecommender(X)
    assert rec.n_threads == 31
    os.environ["IRSPACK_NUM_THREADS_DEFAULT"] = "NOT_A_INTEGER"
    with pytest.raises(ValueError):
        rec = CosineKNNRecommender(X)
    os.environ.pop("IRSPACK_NUM_THREADS_DEFAULT")
    rec = CosineKNNRecommender(X)
    assert rec.n_threads == 1



from typing import Any, List, Optional, Union

import numpy as np
from scipy import sparse as sps

InteractionMatrix = Union[sps.csr_matrix, sps.csc_matrix]
ProfileMatrix = Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]

# wait until better numpy stub support
DenseScoreArray = np.ndarray
DenseMatrix = np.ndarray
UserIndexArray = np.ndarray



import logging
import time
from abc import ABCMeta
from typing import Any, Dict, List, Optional, Tuple, Type

import numpy as np
import optuna
import pandas as pd

from irspack.utils.default_logger import get_default_logger

from ..evaluator import Evaluator
from ..parameter_tuning import Suggestion, overwrite_suggestions
from ..recommenders.base import BaseRecommender, InteractionMatrix
from ..recommenders.base_earlystop import BaseRecommenderWithEarlyStopping


class BaseOptimizer(object, metaclass=ABCMeta):
    """The base optimizer class for recommender classes.

    The child class must define

        - ``recommender_class``
        - ``default_tune_range``

    Args:
        data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The train data.
        val_evaluator (Evaluator):
            The validation evaluator which measures the performance of the recommenders.
        metric (str, optional): Target metric. Defaults to "ndcg".
        logger (Optional[logging.Logger], optional):
            The logger used during the optimization steps. Defaults to None.
            If ``None``, the default logger of irspack will be used.
        suggest_overwrite (List[Suggestion], optional):
            Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
            the default parameter search space defined by ``default_tune_range``
            Defaults to list().
        fixed_params (Dict[str, Any], optional):
            Fixed parameters passed to recommenders during the optimization procedure.
            If such a parameter exists in :obj:`default_tune_range`, it will not be tuned.
            Defaults to dict().

    """

    recommender_class: Type[BaseRecommender]
    default_tune_range: List[Suggestion] = []

    def __init__(
        self,
        data: InteractionMatrix,
        val_evaluator: Evaluator,
        metric: str = "ndcg",
        logger: Optional[logging.Logger] = None,
        suggest_overwrite: List[Suggestion] = list(),
        fixed_params: Dict[str, Any] = dict(),
    ):

        if logger is None:
            logger = get_default_logger()

        self.logger = logger
        self._data = data
        self.val_evaluator = val_evaluator
        self.metric = metric

        self.current_trial: int = 0
        self.best_trial_index: Optional[int] = None
        self.best_val = float("inf")
        self.best_params: Optional[Dict[str, Any]] = None
        self.learnt_config_best: Dict[str, Any] = dict()  # to store early-stopped epoch

        self.valid_results: List[Dict[str, float]] = []
        self.tried_configs: List[Dict[str, Any]] = []
        self.suggestions = overwrite_suggestions(
            self.default_tune_range, suggest_overwrite, fixed_params
        )
        self.fixed_params = fixed_params

    def _suggest(self, trial: optuna.Trial) -> Dict[str, Any]:
        parameters: Dict[str, Any] = dict()
        for s in self.suggestions:
            parameters[s.name] = s.suggest(trial)
        return parameters

    def get_model_arguments(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        return args, kwargs

    def optimize_with_study(
        self,
        study: optuna.Study,
        n_trials: int = 20,
        timeout: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """Perform the optimization step using the user-created ``optuna.Study`` object.
        Creating and managing the study object will be convenient e.g. when you

            1. want to `store/resume the study using RDB backend <https://optuna.readthedocs.io/en/stable/tutorial/003_rdb.html>`_.
            2. want perform a `distributed optimization <https://optuna.readthedocs.io/en/stable/tutorial/004_distributed.html>`_.

        Args:
            study:
                The study object.
            n_trials:
                The number of expected trials (include pruned trial.). Defaults to 20.
            timeout:
                If set to some value (in seconds), the study will exit after that time period.
                Note that the running trials is not interrupted, though. Defaults to None.

        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.

        """
        self.current_trial = -1
        self.best_val = float("inf")
        self.best_time = None
        self.valid_results = []
        self.tried_configs = []

        def objective_func(trial: optuna.Trial) -> float:
            self.current_trial += 1  # for pruning
            start = time.time()
            params = dict(**self._suggest(trial), **self.fixed_params)
            self.logger.info("Trial %s:", self.current_trial)
            self.logger.info("parameter = %s", params)

            arg, parameters = self.get_model_arguments(**params)

            self.tried_configs.append(parameters)
            recommender = self.recommender_class(self._data, *arg, **parameters)
            recommender.learn_with_optimizer(self.val_evaluator, trial)

            score = self.val_evaluator.get_score(recommender)
            end = time.time()

            time_spent = end - start
            score["time"] = time_spent
            self.valid_results.append(score)
            self.logger.info(
                "Config %d obtained the following scores: %s within %f seconds.",
                self.current_trial,
                score,
                time_spent,
            )
            val_score = score[self.metric]
            if (-val_score) < self.best_val:
                self.best_val = -val_score
                self.best_time = time_spent
                self.best_params = parameters
                self.learnt_config_best = dict(**recommender.learnt_config)
                self.logger.info("Found best %s using this config.", self.metric)
                self.best_trial_index = self.current_trial

            return -val_score

        self.logger.info(
            """Start parameter search for %s over the range: %s""",
            type(self).recommender_class.__name__,
            self.suggestions,
        )

        study.optimize(objective_func, n_trials=n_trials, timeout=timeout)
        if self.best_params is None:
            raise RuntimeError("best parameter not found.")
        best_params = dict(**self.best_params)
        best_params.update(**self.learnt_config_best)
        self.best_params = best_params
        result_df = pd.concat(
            [
                pd.DataFrame(self.tried_configs),
                pd.DataFrame(self.valid_results),
            ],
            axis=1,
        ).copy()
        is_best = np.zeros(result_df.shape[0], dtype=np.bool)
        if self.best_trial_index is not None:
            is_best[self.best_trial_index] = True
        result_df["is_best"] = is_best
        return best_params, result_df

    def optimize(
        self,
        n_trials: int = 20,
        timeout: Optional[int] = None,
        random_seed: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """Perform the optimization step.
        ``optuna.Study`` object is created inside this function.

        Args:
            n_trials:
                The number of expected trials (include pruned trial.). Defaults to 20.
            timeout:
                If set to some value (in seconds), the study will exit after that time period.
                Note that the running trials is not interrupted, though. Defaults to None.
            random_seed:
                The random seed to control ``optuna.samplers.TPESampler``. Defaults to None.

        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.

        """
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=random_seed)
        )
        return self.optimize_with_study(study, n_trials, timeout)


class BaseOptimizerWithEarlyStopping(BaseOptimizer):
    """The Base Optimizer class for early-stoppable recommenders.

    The child class must define

        - ``recommender_class``
        - ``default_tune_range``


    Args:
        data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The train data.
        val_evaluator (Evaluator):
            The validation evaluator which measures the performance of the recommenders.
        metric (str, optional): Target metric. Defaults to "ndcg".
        logger (Optional[logging.Logger], optional):
            The logger used during the optimization steps. Defaults to None.
            If ``None``, the default logger of irspack will be used.
        suggest_overwrite (List[Suggestion], optional):
            Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
            the default parameter search space defined by ``default_tune_range``
            Defaults to list().
        fixed_params (Dict[str, Any], optional):
            Fixed parameters passed to recommenders during the optimization procedure.
            If such a parameter exists in ``default_tune_range``, it will not be tuned.
            Defaults to dict().
        max_epoch (int, optional):
            The maximal number of epochs for the training. Defaults to 512.
        validate_epoch (int, optional):
            The frequency of validation score measurement. Defaults to 5.
        score_degradation_max (int, optional):
            Maximal number of allowed score degradation. Defaults to 5. Defaults to 5.
    """

    recommender_class: Type[BaseRecommenderWithEarlyStopping]

    def __init__(
        self,
        data: InteractionMatrix,
        val_evaluator: Evaluator,
        metric: str = "ndcg",
        logger: Optional[logging.Logger] = None,
        suggest_overwrite: List[Suggestion] = list(),
        fixed_params: Dict[str, Any] = dict(),
        max_epoch: int = 512,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        **kwargs: Any,
    ):

        super().__init__(
            data,
            val_evaluator,
            metric,
            logger=logger,
            suggest_overwrite=suggest_overwrite,
            fixed_params=fixed_params,
        )
        self.max_epoch = max_epoch
        self.validate_epoch = validate_epoch
        self.score_degradation_max = score_degradation_max

    def get_model_arguments(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        return super().get_model_arguments(
            *args,
            max_epoch=self.max_epoch,
            validate_epoch=self.validate_epoch,
            score_degradation_max=self.score_degradation_max,
            **kwargs,
        )



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.dataset.movielens import MovieLens100KDataManager
from irspack.split import split_dataframe_partial_user_holdout
from irspack.split.random import UserTrainTestInteractionPair

RNS = np.random.RandomState(0)

df = MovieLens100KDataManager(force_download=True).read_interaction()


def test_user_level_split() -> None:
    dataset, mid_list = split_dataframe_partial_user_holdout(
        df,
        user_column="userId",
        item_column="movieId",
        n_test_user=30,
        n_val_user=30,
        heldout_ratio_val=0.3,
        heldout_ratio_test=0.5,
    )
    train = dataset["train"]
    train_invalid = UserTrainTestInteractionPair(
        train.user_ids, train.X_train[:, :-1], None
    )
    with pytest.raises(ValueError):
        train_invalid.concat(train)
    with pytest.raises(ValueError):
        invalid_arg = UserTrainTestInteractionPair(
            train.user_ids, train.X_train, train.X_train[1:]
        )

    val = dataset["val"]
    test = dataset["test"]
    assert train.X_test.count_nonzero() == 0
    train_val = train.concat(val)
    assert train_val.X_test[: train.n_users].count_nonzero() == 0
    assert (train_val.X_test[train.n_users :] - val.X_test).count_nonzero() == 0

    assert (
        train_val.X_train - sps.vstack([train.X_all, val.X_train])
    ).count_nonzero() == 0

    for user_data, ratio in [(val, 0.3), (test, 0.5)]:
        X_learn = user_data.X_train
        X_predict = user_data.X_test
        assert X_predict is not None
        intersect = X_learn.multiply(X_predict)
        assert intersect.count_nonzero() == 0
        index = RNS.choice(np.arange(user_data.n_users), size=10)
        for i in index:
            nnz_learn = X_learn[i].nonzero()[1].shape[0]
            nnz_predict = X_predict[i].nonzero()[1].shape[0]
            assert ratio >= (nnz_predict - 1) / (nnz_learn + nnz_predict)
            assert ratio <= (nnz_predict + 1) / (nnz_learn + nnz_predict)



from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy import sparse as sps
from sklearn.model_selection import train_test_split

from irspack.definitions import InteractionMatrix
from irspack.utils import rowwise_train_test_split


class UserTrainTestInteractionPair:
    """A class to hold users' train & test (if any) interactions and their ids.

    Args:
        user_ids:
            List of user ids. Its ``i``-th element should correspond to ``i``-th row
            of ``X_train``.
        X_train:
            The train part of interactions.
        X_test:
            The test part of interactions (if any).
            If ``None``, an empty matrix with the shape of ``X_train``
            will be created. Defaults to None.

    Raises:
        ValueError:
            when ``X_train`` and ``user_ids`` have inconsistent size.
        ValueError:
            when ``X_train`` and ``X_test`` have inconsistent size.

    """

    X_train: sps.csr_matrix
    """The train part of users' interactions."""
    X_test: sps.csr_matrix
    """The test part of users' interactions."""
    n_users: int
    """The number of users"""
    n_items: int
    """The number of items"""
    X_all: sps.csr_matrix
    """If ``X_test`` is not ``None``, equal to ``X_train + X_test``.Otherwise equals X_train."""

    def __init__(
        self,
        user_ids: List[Any],
        X_train: InteractionMatrix,
        X_test: Optional[InteractionMatrix],
    ):

        if len(user_ids) != X_train.shape[0]:
            raise ValueError("user_ids and X_train have different shapes.")

        if X_test is not None:
            if X_train.shape != X_test.shape:
                raise ValueError("X_train and X_test have different shapes.")
            X_test = sps.csr_matrix(X_test)
        else:
            X_test = sps.csr_matrix(X_train.shape, dtype=X_train.dtype)
        self.user_ids = [x for x in user_ids]
        self.X_train = sps.csr_matrix(X_train)
        self.X_test = X_test
        self.n_users = self.X_train.shape[0]
        self.n_items = self.X_train.shape[1]
        self.X_all = sps.csr_matrix(self.X_train + self.X_test)

    def concat(
        self, other: "UserTrainTestInteractionPair"
    ) -> "UserTrainTestInteractionPair":
        """Concatenate the users data.
        user_id will be ``self.user_ids + self.item_ids``.

        Returns:
            [type]: [description]

        ValueError:
            when ``self`` and ``other`` have unequal ``n_items``.
        """
        if self.n_items != other.n_items:
            raise ValueError("inconsistent n_items.")
        return UserTrainTestInteractionPair(
            self.user_ids + other.user_ids,
            sps.vstack([self.X_train, other.X_train], format="csr"),
            sps.vstack([self.X_test, other.X_test], format="csr"),
        )


def split_train_test_userwise(
    df_: pd.DataFrame,
    user_colname: str,
    item_colname: str,
    item_id_to_iid: Optional[Dict[Any, int]],
    heldout_ratio: float,
    rns: np.random.RandomState,
    rating_column: Optional[str] = None,
) -> UserTrainTestInteractionPair:
    """Split the user x item data frame into a pair of sparse matrix (represented as a UserDataSet).

    Parameters
    ----------
    df_ : pd.DataFrame
        user x item interaction matrix.
    user_colname : str
        The column name for the users.
    item_colname : str
        The column name for the items.
    item_id_to_iid : Optional[Dict[Any, int]]
        The mapper from item id to item index. If not supplied, create own mapping from df_.
    heldout_ratio : float
        The percentage of items (per-user) to be held out as a test(validation) ones.
    rns : np.random.RandomState
        The random state
    rating_column : Optional[str], optional
        The column for the rating values. If None, the rating values will be all equal (1), by default None

    Returns
    -------
    UserDataSet
        Resulting train-test split dataset.
    """

    if item_id_to_iid is None:
        item_id_to_iid = {id: i for i, id in enumerate(np.unique(df_[item_colname]))}
    df_ = df_[df_[item_colname].isin(item_id_to_iid.keys())]

    item_indices = df_[item_colname].map(item_id_to_iid)

    user_ids, user_indices = np.unique(df_[user_colname], return_inverse=True)
    if rating_column is not None:
        data = df_[rating_column].values
    else:
        data = np.ones(df_.shape[0], dtype=np.int32)

    X_all = sps.csr_matrix(
        (data, (user_indices, item_indices)),
        shape=(len(user_ids), len(item_id_to_iid)),
    )
    X_learn, X_predict = rowwise_train_test_split(
        X_all,
        heldout_ratio,
        random_seed=rns.randint(-(2 ** 31), 2 ** 31 - 1),
    )

    return UserTrainTestInteractionPair(user_ids, X_learn.tocsr(), X_predict.tocsr())


def split_dataframe_partial_user_holdout(
    df_all: pd.DataFrame,
    user_column: str,
    item_column: str,
    rating_column: Optional[str] = None,
    n_val_user: Optional[int] = None,
    n_test_user: Optional[int] = None,
    val_user_ratio: float = 0.1,
    test_user_ratio: float = 0.1,
    heldout_ratio_val: float = 0.5,
    heldout_ratio_test: float = 0.5,
    random_state: int = 42,
) -> Tuple[Dict[str, UserTrainTestInteractionPair], List[Any]]:
    """Splits the DataFrame and build an interaction matrix,
    holding out random interactions for a subset of randomly selected users
    (whom we call "validation users" and "test users").

    Args:
        df_all:
            The user-item interaction event log.
        user_column:
            The column name for user_id.
        item_column:
            The column name for movie_id.
        rating_column:
            The column name for ratings. If ``None``, the rating will be treated as
            ``1`` for all interactions. Defaults to None.
        n_val_user:
            The number of "validation users". Defaults to None.
        n_test_user:
            The number of "test users". Defaults to None.
        val_user_ratio:
            The percentage of "validation users" with respect to all users.
            Ignored when ``n_val_user`` is set. Defaults to 0.1.
        test_user_ratio:
            The percentage of "test users" with respect to all users.
            Ignored when ``n_text_user`` is set. Defaults to 0.1.
        heldout_ratio_val:
            The percentage of held-out interactions for "validation users". Defaults to 0.5.
        heldout_ratio_test:
            The percentage of held-out interactions for "test users". Defaults to 0.5.
        random_state:
            The random seed for this procedure. Defaults to 42.

    Raises:
        ValueError: When ``n_val_user + n_test_user`` is greater than the number of total users.

    Returns:
        A tuple consisting of:

            1. A dictionary with ``"train"``, ``"val"``, ``"test"`` as its keys and the
               coressponding dataset as its values.
            2. List of unique item ids (which corresponds to the columns of the datasets).
    """
    assert (test_user_ratio <= 1) and (test_user_ratio >= 0)
    assert (val_user_ratio <= 1) and (val_user_ratio >= 0)

    uids: List[Any] = df_all[user_column].unique()
    n_users_all = len(uids)
    if n_val_user is None:
        n_val_user = int(n_users_all * val_user_ratio)
        val_user_ratio = n_val_user / n_users_all
    else:
        if n_val_user > n_users_all:
            raise ValueError("n_val_user exceeds the number of total users.")

    if n_test_user is None:
        n_test_user = int(n_users_all * test_user_ratio)
        test_user_ratio = n_test_user / n_users_all
    else:
        if (n_test_user + n_val_user) > n_users_all:
            raise ValueError(
                "n_val_user + n_test_users exceeds the number of total users."
            )

    df_all = df_all.drop_duplicates([user_column, item_column])
    rns = np.random.RandomState(random_state)
    train_uids, val_test_uids = train_test_split(
        uids, test_size=(n_val_user + n_test_user), random_state=rns
    )

    if (test_user_ratio * len(uids)) >= 1:
        val_uids, test_uids = train_test_split(
            val_test_uids,
            test_size=n_test_user,
            random_state=rns,
        )
    else:
        val_uids = val_test_uids
        test_uids = np.asarray([], dtype=val_uids.dtype)
    df_train = df_all[df_all[user_column].isin(train_uids)].copy()
    df_val = df_all[df_all[user_column].isin(val_uids)]
    df_test = df_all[df_all[user_column].isin(test_uids)]
    item_all: List[Any] = list(df_train[item_column].unique())

    def select_item(df: pd.DataFrame) -> pd.DataFrame:
        return df[df[item_column].isin(item_all)]

    df_val = select_item(df_val).copy()
    df_test = select_item(df_test).copy()
    item_id_to_iid = {id_: i for i, id_ in enumerate(item_all)}
    for df in [df_train, df_val, df_test]:
        df["item_iid"] = df[item_column].map(item_id_to_iid)

    train_uids = df_train[user_column].unique()
    train_uid_to_iid = {uid: iid for iid, uid in enumerate(train_uids)}

    train_user_row = df_train[user_column].map(train_uid_to_iid).values
    train_user_col = df_train.item_iid
    if rating_column is None:
        train_user_data = np.ones(df_train.shape[0])
    else:
        train_user_data = df_train[rating_column].values

    train_user_interactions = sps.csr_matrix(
        (train_user_data, (train_user_row, train_user_col)),
        shape=(len(train_uids), len(item_id_to_iid)),
    )

    valid_data: Dict[str, UserTrainTestInteractionPair] = dict(
        train=UserTrainTestInteractionPair(train_uids, train_user_interactions, None)
    )

    for df_, dataset_name, heldout_ratio in [
        (df_val, "val", heldout_ratio_val),
        (df_test, "test", heldout_ratio_test),
    ]:

        valid_data[dataset_name] = split_train_test_userwise(
            df_,
            user_column,
            item_column,
            item_id_to_iid,
            heldout_ratio,
            rns,
            rating_column=rating_column,
        )
    return valid_data, item_all



import json
import os
from typing import Any, Dict, List, Tuple, Type

import pandas as pd
from scipy import sparse as sps

from irspack.dataset.movielens import MovieLens20MDataManager
from irspack.evaluator import EvaluatorWithColdUser
from irspack.optimizers import (
    AsymmetricCosineKNNOptimizer,
    BaseOptimizer,
    CosineKNNOptimizer,
    DenseSLIMOptimizer,
    IALSOptimizer,
    MultVAEOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    SLIMOptimizer,
    TopPopOptimizer,
)
from irspack.split import split_dataframe_partial_user_holdout

N_CPUS = os.cpu_count()
if N_CPUS is None:
    N_CPUS = 1
os.environ["OMP_NUM_THREADS"] = str(N_CPUS)

# This will set the number of thread to be N_CPUS where it is possible.
# You can also controll the number of threads for each recommender.
os.environ["IRSPACK_NUM_THREADS_DEFAULT"] = str(N_CPUS)

if __name__ == "__main__":

    BASE_CUTOFF = 100

    # We follow the preprocessing of Mult-VAE implementation (https://github.com/dawenl/vae_cf)
    data_manager = MovieLens20MDataManager()
    df_all = data_manager.read_interaction()
    df_all = df_all[df_all.rating >= 4]
    user_cnt = df_all.userId.value_counts()
    user_cnt = user_cnt[user_cnt >= 5]
    df_all = df_all[df_all.userId.isin(user_cnt.index)]

    data_all, _ = split_dataframe_partial_user_holdout(
        df_all,
        "userId",
        "movieId",
        n_test_user=10000,
        n_val_user=10000,
        heldout_ratio_val=0.2,
        heldout_ratio_test=0.2,
    )

    data_train = data_all["train"]
    data_val = data_all["val"]
    data_test = data_all["test"]

    X_train_val_all: sps.csr_matrix = sps.vstack(
        [data_train.X_all, data_val.X_all], format="csr"
    )
    valid_evaluator = EvaluatorWithColdUser(
        input_interaction=data_val.X_train,
        ground_truth=data_val.X_test,
        cutoff=BASE_CUTOFF,
    )
    test_evaluator = EvaluatorWithColdUser(
        input_interaction=data_test.X_train,
        ground_truth=data_test.X_test,
        cutoff=BASE_CUTOFF,
    )

    test_results = []
    validation_results = []

    test_configs: List[Tuple[Type[BaseOptimizer], int, Dict[str, Any]]] = [
        (TopPopOptimizer, 1, dict()),
        (CosineKNNOptimizer, 40, dict()),
        (AsymmetricCosineKNNOptimizer, 40, dict()),
        (P3alphaOptimizer, 30, dict(alpha=1)),
        (RP3betaOptimizer, 40, dict(alpha=1)),
        (IALSOptimizer, 40, dict()),
        (DenseSLIMOptimizer, 20, dict()),
        (
            MultVAEOptimizer,
            1,
            dict(
                dim_z=200, enc_hidden_dims=600, kl_anneal_goal=0.2
            ),  # nothing to tune, use the parameters used in the paper.
        ),
        # (SLIMOptimizer, 40, dict()), # Note: this is a heavy one.
    ]
    for optimizer_class, n_trials, config in test_configs:
        recommender_name = optimizer_class.recommender_class.__name__
        optimizer: BaseOptimizer = optimizer_class(
            data_train.X_all,
            valid_evaluator,
            metric="ndcg",
            fixed_params=config,
        )
        (best_param, validation_result_df) = optimizer.optimize(n_trials=n_trials)
        validation_result_df["recommender_name"] = recommender_name
        validation_results.append(validation_result_df)
        pd.concat(validation_results).to_csv(f"validation_scores.csv")
        test_recommender = optimizer.recommender_class(X_train_val_all, **best_param)
        test_recommender.learn()
        test_scores = test_evaluator.get_scores(test_recommender, [20, 50, 100])

        test_results.append(
            dict(name=recommender_name, best_param=best_param, **test_scores)
        )
        with open("test_results.json", "w") as ofs:
            json.dump(test_results, ofs, indent=2)



from typing import Dict

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.recommenders import IALSRecommender

X_small = sps.csr_matrix(
    np.asfarray([[6, 1, 2, 3, 4], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 1]])
)


def test_ials_overfit_cholesky(
    test_interaction_data: Dict[str, sps.csr_matrix]
) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(X, n_components=4, alpha=0, reg=0, use_cg=False)
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.trainer.core_trainer.transform_user(X.tocsr().astype(np.float32))
    ivec = rec.trainer.core_trainer.transform_item(X.tocsr().astype(np.float32))
    X = X.toarray()
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X, rtol=1e-2, atol=1e-2)


def test_ials_overfit_cg(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X, n_components=4, alpha=0, reg=1e-2, use_cg=True, max_cg_steps=4
    )
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.trainer.core_trainer.transform_user(X.tocsr().astype(np.float32))
    ivec = rec.trainer.core_trainer.transform_item(X.tocsr().astype(np.float32))
    X_dense = X.toarray()
    reproduced_user_vector = uvec.dot(ivec.T)
    np.testing.assert_allclose(reproduced_user_vector, X_dense, rtol=1e-2, atol=1e-2)

    X_reproduced = rec.get_score_cold_user(X)
    np.testing.assert_allclose(X_reproduced, X_dense, rtol=1e-2, atol=1e-2)

    with pytest.raises(ValueError):
        _ = rec.compute_item_embedding(X.T)

    reproduced_item_vector = rec.compute_item_embedding(X)
    X_reproduced_item = rec.get_score_from_item_embedding(
        np.arange(X.shape[0]), reproduced_item_vector
    )
    np.testing.assert_allclose(X_reproduced_item, X_dense, rtol=1e-2, atol=1e-2)


@pytest.mark.xfail
def test_ials_cg_underfit(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    X = test_interaction_data["X_small"]
    rec = IALSRecommender(
        X,
        n_components=4,
        alpha=0,
        reg=1e-2,
        use_cg=True,
        max_cg_steps=1,
        max_epoch=10,
    )
    with pytest.raises(RuntimeError):
        _ = rec.core_trainer.user
    rec.learn()
    assert rec.trainer is not None
    uvec = rec.core_trainer.user
    ivec = rec.core_trainer.item
    X_dense = X.toarray()
    reprod = uvec.dot(ivec.T)
    np.testing.assert_allclose(reprod, X_dense, rtol=1e-2, atol=1e-2)



import uuid
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy import sparse as sps
from sklearn.model_selection import train_test_split

from irspack.definitions import InteractionMatrix

from .random import UserTrainTestInteractionPair


def holdout_specific_interactions(
    df: pd.DataFrame,
    user_column: str,
    item_column: str,
    interaction_indicator: np.ndarray,
    validatable_user_ratio_val: float = 0.2,
    validatable_user_ratio_test: float = 0.2,
    random_seed: Optional[int] = None,
) -> Tuple[List[Any], Dict[str, UserTrainTestInteractionPair]]:
    """Holds-out (part of) the interactions specified by the users.

    All the users will be split into two category:

        1. Those who have an interaction in the specified subset.
           We denote them as  "validatable" users.
        2. Those who don't.

    We split the users in 1. into three parts (train, validation, test)-users, and hold-out the specified interactions.
    The interactions of non-validatable users will be part of the train dataset.

    This split will be useful when want to:

        - recommend only part of the items (e.g., rather unpopular ones) to the users.
          In this case, the held-out interactions will be the ones with these specific items.
        - split the dataframe by a certain timepoint, and ensure that no information after
          that timepoint contaminates the training set.

    Args:
        df:
            The data source.
        user_column:
            The column name of the users.
        item_column:
            The column name of the items.
        interaction_indicator:
            Specifies where in ``df`` the held-out interactions are.
        validatable_user_ratio_val:
            The ratio of "validation-set users" in the "validatable users". Defaults to 0.2.
        validatable_user_ration_test:
            The ratio of "test-set users" in the "validatable users". Defaults to 0.2.
        random_seed:
            THe random seed used to split validatable users into three. Defaults to None.

    Returns:
        A tuple consiting of

            * The aligned list of all the items.
            * A dictionary with train/val/test user pairs.
    """
    v_user_ratio_train = 1 - validatable_user_ratio_val - validatable_user_ratio_test

    if v_user_ratio_train < -1e-10:
        raise ValueError(
            "validatable_use_ratio_val + validatable_user_ratio_test exceeds 1."
        )

    df = df[[user_column, item_column]].copy()

    unique_item_ids, item_index = np.unique(df[item_column], return_inverse=True)
    item_index_colname = str(uuid.uuid1())
    df[item_index_colname] = item_index

    flg_colname = str(uuid.uuid1())
    flg_column = np.zeros(df.shape[0])
    flg_column[interaction_indicator] = 1
    df[flg_colname] = flg_column

    validatable_users = np.unique(df[flg_column > 0][user_column])
    val_test_ratio = 1 - v_user_ratio_train
    if val_test_ratio >= 1.0:
        v_train_users = np.ndarray((0,), dtype=validatable_users.dtype)
        v_val_test_users = validatable_users
    else:
        v_train_users, v_val_test_users = train_test_split(
            validatable_users,
            test_size=val_test_ratio,
            random_state=random_seed,
        )
    v_val_users, v_test_users = train_test_split(
        v_val_test_users,
        test_size=(validatable_user_ratio_test) / (1 - v_user_ratio_train),
        random_state=random_seed,
    )
    df_train = pd.concat(
        [
            df[df[user_column].isin(v_train_users)],
            df[~df[user_column].isin(validatable_users)],
        ]
    )
    df_val = df[df[user_column].isin(v_val_users)]
    df_test = df[df[user_column].isin(v_test_users)]

    def df_to_dataset(
        df: pd.DataFrame, train: bool = False
    ) -> UserTrainTestInteractionPair:
        uid_unique, uindex = np.unique(df[user_column], return_inverse=True)
        iindex = df[item_index_colname].values
        if train:
            X = sps.csr_matrix(
                (np.ones(df.shape[0], dtype=np.float64), (uindex, iindex)),
                shape=(len(uid_unique), len(unique_item_ids)),
            )
            return UserTrainTestInteractionPair(uid_unique, X_train=X, X_test=None)
        (learn_index,) = np.where((df[flg_colname] == 0).values)
        (predict_index,) = np.where((df[flg_colname] > 0).values)
        X_learn = sps.csr_matrix(
            (
                np.ones(learn_index.shape[0], dtype=np.float64),
                (uindex[learn_index], iindex[learn_index]),
            ),
            shape=(len(uid_unique), len(unique_item_ids)),
        )
        X_predict = sps.csr_matrix(
            (
                np.ones(predict_index.shape[0], dtype=np.float64),
                (uindex[predict_index], iindex[predict_index]),
            ),
            shape=(len(uid_unique), len(unique_item_ids)),
        )
        return UserTrainTestInteractionPair(
            uid_unique, X_train=X_learn, X_test=X_predict
        )

    dataset = {
        "train": df_to_dataset(df_train, train=True),
        "val": df_to_dataset(df_val, train=False),
        "test": df_to_dataset(df_test, train=False),
    }
    return unique_item_ids, dataset



import warnings

from .base import BaseRecommender, BaseSimilarityRecommender
from .base_earlystop import BaseRecommenderWithEarlyStopping
from .dense_slim import DenseSLIMRecommender
from .ials import IALSRecommender
from .knn import (
    AsymmetricCosineKNNRecommender,
    CosineKNNRecommender,
    JaccardKNNRecommender,
    TverskyIndexKNNRecommender,
)
from .nmf import NMFRecommender
from .p3 import P3alphaRecommender
from .rp3 import RP3betaRecommender
from .rwr import RandomWalkWithRestartRecommender
from .slim import SLIMRecommender
from .toppop import TopPopRecommender
from .truncsvd import TruncatedSVDRecommender

__all__ = [
    "BaseRecommender",
    "BaseSimilarityRecommender",
    "BaseRecommenderWithEarlyStopping",
    "TopPopRecommender",
    "P3alphaRecommender",
    "RP3betaRecommender",
    "DenseSLIMRecommender",
    "NMFRecommender",
    "RandomWalkWithRestartRecommender",
    "SLIMRecommender",
    "TruncatedSVDRecommender",
    "IALSRecommender",
    "CosineKNNRecommender",
    "JaccardKNNRecommender",
    "TverskyIndexKNNRecommender",
    "AsymmetricCosineKNNRecommender",
]

try:
    from .multvae import MultVAERecommender

    __all__.append("MultVAERecommender")
except ModuleNotFoundError:
    warnings.warn("Failed to import MultVAERecommender")

try:
    from .bpr import BPRFMRecommender

    __all__.append("BPRFMRecommender")
except:
    warnings.warn(
        "Failed to import BPRFMRecommender. To enable this feature, install `lightfm`."
    )
    pass



import pickle
from dataclasses import dataclass
from typing import IO, Any, Dict, List, Optional, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
from optax import OptState, adam
from scipy import sparse as sps

from ..definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from .base_earlystop import BaseRecommenderWithEarlyStopping, TrainerBase


class BaseMLP:
    def __init__(
        self,
        output_dim: int,
        hidden_dims: List[int],
    ):
        self.output_dim = output_dim
        self.hidden_dims = hidden_dims

    def __call__(
        self,
        X: jnp.ndarray,
    ) -> jnp.ndarray:
        layers = []
        for d_o in self.hidden_dims:
            layers.append(hk.Linear(d_o))
            layers.append(jnp.tanh)
        layers.append(hk.Linear(self.output_dim))

        return hk.Sequential(layers)(X)


class DecoderNN:
    def __init__(self, output_dim: int, hidden_dims: List[int]):
        mlp = BaseMLP(output_dim=output_dim, hidden_dims=hidden_dims)
        self.mlp = mlp

    def __call__(self, X: jnp.ndarray) -> jnp.ndarray:
        return jax.nn.log_softmax(self.mlp(X), axis=1)


def l2_normalize(X: jnp.ndarray) -> jnp.ndarray:
    return X / jnp.sqrt((X ** 2).sum(axis=1) + 1e-8)[:, None]


class EncoderNN:
    def __init__(
        self,
        latent_dim: int,
        hidden_dims: List[int],
    ):
        self.mlp = BaseMLP(2 * latent_dim, hidden_dims)
        self.latent_dim = latent_dim

    def __call__(
        self, X: jnp.ndarray, dropout: float, train: bool
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        X = l2_normalize(X)
        if train:
            X = hk.dropout(hk.next_rng_key(), dropout, X)
        h = self.mlp(X)
        mu = h[:, : self.latent_dim]
        log_var = h[:, self.latent_dim :]
        return mu, log_var


@dataclass
class MultVAEOutput:
    log_softmax: jnp.ndarray
    mean: jnp.ndarray
    log_stddev: jnp.ndarray
    KL: jnp.ndarray


class MultVAE:
    def __init__(
        self,
        n_obs: int,
        latent_dim: int,
        enc_hidden_dims: List[int],
        dec_hidden_dims: List[int],
        dropout_p: float = 0.5,
        l2_reg: float = 0.01,
    ):
        self.encoder_network = EncoderNN(
            latent_dim,
            enc_hidden_dims,
        )
        self.decoder_network = DecoderNN(n_obs, dec_hidden_dims)
        self._kl_coeff: float = 0.0
        self.dropout_p = dropout_p
        self.l2_reg = l2_reg

    def set_kl_coeff(self, nv: float) -> None:
        self._kl_coeff = nv

    def kl_coeff(self) -> float:
        return self._kl_coeff

    def __call__(self, X: jnp.ndarray, p: jnp.ndarray, train: bool) -> MultVAEOutput:
        mu, log_var = self.encoder_network(X, p, train)
        std = jnp.exp(log_var * 0.5)
        # log_var = 2 * log (std)
        KL: jnp.ndarray = 0.5 * (-log_var + std ** 2 + mu ** 2 - 1)
        KL = KL.sum(axis=1).mean(axis=0)

        if train:
            eps = jax.random.normal(hk.next_rng_key(), mu.shape)  # self.rng.rand
            z: jnp.ndarray = mu + eps * std
        else:
            z = mu
        log_softmax: jnp.ndarray = self.decoder_network(z)

        return MultVAEOutput(log_softmax, mu, log_var * 0.5, KL)


class MultVAETrainer(TrainerBase):
    def __init__(
        self,
        X: InteractionMatrix,
        dim_z: int,
        enc_hidden_dim: int,
        dec_hidden_dim: Optional[int],
        dropout_p: float,
        l2_regularizer: float,
        kl_anneal_goal: float,
        anneal_end_epoch: int,
        minibatch_size: int,
        learning_rate: float,
    ):
        self.X = X
        self.n_users = X.shape[0]
        self.n_items = X.shape[1]
        self.minibatch_size = minibatch_size
        self.kl_anneal_goal = kl_anneal_goal
        self.anneal_end_epoch = anneal_end_epoch
        self.dropout_p = dropout_p
        self.l2_regularizer = l2_regularizer
        self.learning_rate = learning_rate

        self.rng_seq = hk.PRNGSequence(42)

        if dec_hidden_dim is None:
            dec_hidden_dim = enc_hidden_dim

        self.enc_hidden_dim = enc_hidden_dim
        self.dec_hidden_dim = dec_hidden_dim
        self.dim_z = dim_z

        self.total_anneal_step = (anneal_end_epoch * self.n_users) / minibatch_size
        self._setup_jax_funcs()

        self._update_count = 0
        params = self.vae_f.init(
            next(self.rng_seq),
            np.zeros((1, self.n_items), dtype=np.float32),
            self.dropout_p,
            True,
        )

        self.params = params

        self.opt_state = self.optimizer.init(self.params)

    def _setup_jax_funcs(self) -> None:
        vae_f = hk.transform(
            lambda X, p, train: MultVAE(
                self.n_items,
                self.dim_z,
                [self.enc_hidden_dim],
                [self.dec_hidden_dim],
                dropout_p=self.dropout_p,
                l2_reg=self.l2_regularizer,
            )(X, p, train)
        )
        self.optimizer = adam(learning_rate=self.learning_rate)

        def loss_fn(
            params: hk.Params,
            rng: jnp.ndarray,
            X: jnp.ndarray,
            kl_coeff: jnp.ndarray,
            dropout: float,
            train: bool,
        ) -> jnp.ndarray:
            mvresult: MultVAEOutput = vae_f.apply(params, rng, X, dropout, train)
            neg_ll = -(mvresult.log_softmax * X).sum(axis=1).mean()
            neg_elbo = neg_ll + kl_coeff * mvresult.KL
            return neg_elbo

        loss_fn = jax.jit(loss_fn, static_argnums=(4, 5))

        def update(
            params: hk.Params,
            rng: jnp.ndarray,
            opt_state: OptState,
            X: jnp.ndarray,
            kl_coeff: jnp.ndarray,
            dropout: float,
        ) -> Tuple[hk.Params, OptState]:
            grads = jax.grad(loss_fn)(params, rng, X, kl_coeff, dropout, True)
            updates, new_optstate = self.optimizer.update(grads, opt_state)
            new_params = optax.apply_updates(
                params,
                updates,
            )
            return new_params, new_optstate

        update = jax.jit(update, static_argnums=(5,))
        self.vae_f = vae_f
        self.update_function = update

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        mb_arrays: List[DenseScoreArray] = []
        n_users: int = X.shape[0]
        X_csr: sps.csr_matrix = X.tocsr()
        for mb_start in range(0, n_users, self.minibatch_size):
            mb_end = min(n_users, mb_start + self.minibatch_size)
            X_mb = X_csr[mb_start:mb_end].astype(np.float32).toarray()
            X_jax = jnp.asarray(X_mb)
            mvresult: MultVAEOutput = self.vae_f.apply(
                self.params, next(self.rng_seq), X_jax, self.dropout_p, False
            )

            mb_arrays.append(np.asarray(mvresult.log_softmax, dtype=np.float64))
        score_concat = np.concatenate(mb_arrays, axis=0)
        return score_concat

    def run_epoch(self) -> None:
        user_indices = np.arange(self.n_users)
        np.random.shuffle(user_indices)
        for mb_start in range(0, self.n_users, self.minibatch_size):
            current_kl_coeff = jnp.asarray(
                self.kl_anneal_goal
                * min(1, self._update_count / self.total_anneal_step)
            )
            mb_end = min(self.n_users, mb_start + self.minibatch_size)
            X = self.X[user_indices[mb_start:mb_end]].astype(np.float32)
            if sps.issparse(X):
                X = X.toarray()
            X_jax = jnp.asarray(X, dtype=jnp.float32)
            self.params, self.opt_state = self.update_function(
                self.params,
                next(self.rng_seq),
                self.opt_state,
                X_jax,
                current_kl_coeff,
                self.dropout_p,
            )
            self._update_count += 1

    def save_state(self, ofs: IO) -> None:
        pickle.dump(self.params, ofs)

    def load_state(self, ifs: IO) -> None:
        self.params = pickle.load(ifs)

    def __getstate__(self) -> Any:
        serealized = dict(**self.__dict__)
        serealized.pop("update_function", None)
        serealized.pop("vae_f", None)
        serealized.pop("optimizer")
        return serealized

    def __setstate__(self, state: Dict[str, Any]) -> Any:
        self.__dict__.update(state)
        self._setup_jax_funcs()


class MultVAERecommender(BaseRecommenderWithEarlyStopping):
    def __init__(
        self,
        X_train_all: InteractionMatrix,
        dim_z: int = 16,
        enc_hidden_dims: int = 256,
        dec_hidden_dims: Optional[int] = None,
        dropout_p: float = 0.5,
        l2_regularizer: float = 0,
        kl_anneal_goal: float = 0.2,
        anneal_end_epoch: int = 50,
        minibatch_size: int = 512,
        max_epoch: int = 300,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        learning_rate: float = 1e-3,
    ) -> None:
        super().__init__(
            X_train_all,
            max_epoch=max_epoch,
            validate_epoch=validate_epoch,
            score_degradation_max=score_degradation_max,
        )

        self.dim_z = dim_z
        self.enc_hidden_dims = enc_hidden_dims
        self.dec_hidden_dims = dec_hidden_dims
        self.kl_anneal_goal = kl_anneal_goal
        self.anneal_end_epoch = anneal_end_epoch
        self.minibatch_size = minibatch_size
        self.dropout_p = dropout_p
        self.l2_regularizer = l2_regularizer
        self.learning_rate = learning_rate

        self.trainer: Optional[MultVAETrainer] = None

    def _create_trainer(self) -> MultVAETrainer:
        return MultVAETrainer(
            self.X_train_all,
            self.dim_z,
            self.enc_hidden_dims,
            self.dec_hidden_dims,
            self.dropout_p,
            self.l2_regularizer,
            self.kl_anneal_goal,
            self.anneal_end_epoch,
            self.minibatch_size,
            self.learning_rate,
        )

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.get_score_cold_user(self.X_train_all[user_indices])

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        if self.trainer is None:
            raise RuntimeError("encoder called before training.")
        return self.trainer.get_score_cold_user(X)



import warnings

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.utils import (
    okapi_BM_25_weight,
    rowwise_train_test_split,
    sparse_mm_threaded,
    tf_idf_weight,
)

X = sps.csr_matrix(
    np.asfarray([[1, 1, 2, 3, 4], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]])
)
X.sort_indices()
X_array = X.toarray()


def test_sparse_mm_threaded() -> None:
    X_mm = sparse_mm_threaded(X, X.T, 3).toarray()
    X_sps = X.dot(X.T).toarray()
    X_diff = X_mm - X_sps
    assert np.all(X_diff == 0)


def test_split() -> None:
    warnings.simplefilter("always")

    X_1, X_2 = rowwise_train_test_split(X, test_ratio=0.5, random_seed=1)
    assert np.all((X - X_1 - X_2).toarray() == 0)

    # should have no overwrap
    assert np.all(X_1.multiply(X_2).toarray() == 0)


def test_bm25() -> None:
    k1 = 1.4
    b = 0.8
    X_weighted = okapi_BM_25_weight(X, k1=k1, b=b).toarray()

    df = (X.toarray() > 0).sum(axis=0).ravel()
    idf = np.log(X.shape[0] / (df + 1) + 1)
    avgdl = X_array.sum(axis=1).mean()
    for row in range(X.shape[0]):
        for col in range(X.shape[1]):
            tf = X_array[row, col]
            py_answer = (
                idf[col]
                * (tf * (k1 + 1))
                / (tf + k1 * (1 - b + b * X_array[row].sum() / avgdl))
            )
            assert py_answer == pytest.approx(X_weighted[row, col])


def test_tf_idf() -> None:
    X_manual = (
        X.toarray() * np.log(X.shape[0] / (1 + np.bincount(X.nonzero()[1])))[None, :]
    )
    X_tf_idf = tf_idf_weight(X).toarray()
    assert np.all((X_manual - X_tf_idf) == 0.0)



from typing import Iterable as iterable
from typing import Iterator as iterator
from typing import *

from numpy import float64

import irspack.recommenders._knn

_Shape = Tuple[int, ...]
import scipy.sparse

__all__ = [
    "AsymmetricSimilarityComputer",
    "CosineSimilarityComputer",
    "JaccardSimilarityComputer",
    "P3alphaComputer",
    "RP3betaComputer",
    "TverskyIndexComputer",
]

class AsymmetricSimilarityComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        shrinkage: float,
        alpha: float,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_similarity(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csr_matrix[float64]: ...
    pass

class CosineSimilarityComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        shrinkage: float,
        normalize: bool,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_similarity(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csr_matrix[float64]: ...
    pass

class JaccardSimilarityComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        shrinkage: float,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_similarity(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csr_matrix[float64]: ...
    pass

class P3alphaComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        alpha: float = 0,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_W(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csc_matrix[float64]: ...
    pass

class RP3betaComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        alpha: float = 0,
        beta: float = 0,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_W(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csc_matrix[float64]: ...
    pass

class TverskyIndexComputer:
    def __init__(
        self,
        X: scipy.sparse.csr_matrix[float64],
        shrinkage: float,
        alpha: float,
        beta: float,
        n_threads: int = 1,
        max_chunk_size: int = 128,
    ) -> None: ...
    def compute_similarity(
        self, arg0: scipy.sparse.csr_matrix[float64], arg1: int
    ) -> scipy.sparse.csr_matrix[float64]: ...
    pass



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.dataset.movielens import MovieLens100KDataManager
from irspack.definitions import DenseScoreArray, UserIndexArray
from irspack.evaluator import Evaluator
from irspack.optimizers import BaseOptimizer
from irspack.parameter_tuning import (
    CategoricalSuggestion,
    IntegerLogUniformSuggestion,
    IntegerSuggestion,
    LogUniformSuggestion,
    UniformSuggestion,
)
from irspack.recommenders import BaseRecommender
from irspack.split import rowwise_train_test_split

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(100, 32) > 0.8).astype(np.float64)
)
ml_100k_df = MovieLens100KDataManager(force_download=True).read_interaction()
_, user_index = np.unique(ml_100k_df.userId, return_inverse=True)
_, movie_index = np.unique(ml_100k_df.movieId, return_inverse=True)
X_large = sps.csr_matrix(
    (np.ones(ml_100k_df.shape[0], dtype=np.float64), (user_index, movie_index)),
)


class MockRecommender(BaseRecommender):
    def __init__(
        self,
        X: sps.csr_matrix,
        X_test: sps.csr_matrix,
        p1: float = 1,
        I1: int = 1,
        I2: int = 1,
        reg: float = 1.0,
        flag: str = "hoge",
    ):
        super().__init__(X)
        self.p1 = p1  # only p1 matters
        self.I1 = I1
        self.I2 = I2
        self.reg = reg
        self.answer = X_test
        self.flag = flag
        self.rns = np.random.RandomState(0)

    def _learn(self) -> None:
        pass

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        score = self.answer[user_indices] * self.p1
        score = score + 10 * self.rns.rand(*score.shape) * (1 - self.p1)
        return score


class MockOptimizer(BaseOptimizer):
    recommender_class = MockRecommender
    default_tune_range = [
        UniformSuggestion("p1", 0, 1),
        LogUniformSuggestion("reg", 0.99, 1.01),
        IntegerSuggestion("I1", 100, 102),
        IntegerLogUniformSuggestion("I2", 500, 502),
        CategoricalSuggestion("flag", ["foo", "bar"]),
    ]


@pytest.mark.parametrize("X", [X_small, X_large])
def test_optimizer_by_mock(X: sps.csr_matrix) -> None:
    X_train, X_val = rowwise_train_test_split(X, test_ratio=0.5, random_seed=0)
    evaluator = Evaluator(X_val, 0)
    optimizer = MockOptimizer(
        X_train, evaluator, logger=None, fixed_params=dict(X_test=X_val)
    )
    config, _ = optimizer.optimize(n_trials=40, random_seed=42)
    assert config["p1"] >= 0.9
    assert (config["reg"] >= 0.99) and (config["reg"] <= 1.01)
    assert (config["I1"] >= 100) and (config["I1"] <= 102)
    assert (config["I2"] >= 500) and (config["I2"] <= 502)
    assert config["flag"] in ["foo", "bar"]



m: int
n: int
from typing import Iterable as iterable
from typing import Iterator as iterator
from typing import *

from numpy import float32, float64

import irspack._evaluator

_Shape = Tuple[int, ...]
import flags
import numpy
import scipy.sparse

__all__ = ["EvaluatorCore", "Metrics"]

class EvaluatorCore:
    def __init__(
        self,
        grount_truth: scipy.sparse.csr_matrix[float64],
        recommendable: List[List[int]],
    ) -> None: ...
    def get_ground_truth(self) -> scipy.sparse.csr_matrix[float64]: ...
    def get_metrics(
        self,
        score_array: numpy.ndarray[float64[m, n], flags.writeable, flags.c_contiguous],
        cutoff: int,
        offset: int,
        n_threads: int,
        recall_with_cutoff: bool = False,
    ) -> Metrics: ...
    pass

class Metrics:
    def __init__(self, arg0: int) -> None: ...
    def as_dict(self) -> Dict[str, float]: ...
    def merge(self, arg0: Metrics) -> None: ...
    pass



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List

from optuna import Trial

__all__ = [
    "Suggestion",
    "UniformSuggestion",
    "LogUniformSuggestion",
    "IntegerSuggestion",
    "IntegerLogUniformSuggestion",
    "CategoricalSuggestion",
    "overwrite_suggestions",
]


class Suggestion(object, metaclass=ABCMeta):
    def __init__(self, name: str):
        """The base class to controll optuna's ``Trial`` behavior during
            hyper parameter optimization.

        Args:
            name (str): The name of the parameter to be tuned.
        """
        self.name = name

    @abstractmethod
    def suggest(self, trial: Trial) -> Any:
        raise NotImplementedError('"suggest" must be implemented.')

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError('"__repr__" must be implemented.')


def overwrite_suggestions(
    base: List[Suggestion],
    suggest_overwrite: List[Suggestion],
    fixed_params: Dict[str, Any],
) -> List[Suggestion]:
    for suggest in suggest_overwrite:
        if suggest.name in fixed_params:
            raise ValueError("suggest_overwrite and fixed_param have overwrap.")

    overwritten_parameter_names = set(
        [x.name for x in suggest_overwrite] + [x for x in fixed_params]
    )
    suggestions = [
        suggest_base
        for suggest_base in base
        if suggest_base.name not in overwritten_parameter_names
    ] + suggest_overwrite
    return suggestions


class UniformSuggestion(Suggestion):
    def __init__(self, name: str, low: float, high: float):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: Trial) -> Any:
        return trial.suggest_uniform(self.name, self.low, self.high)

    def __repr__(self) -> str:
        return f"UniformSuggestion(name={self.name!r}, low={self.low!r}, high={self.high!r})"


class LogUniformSuggestion(Suggestion):
    def __init__(self, name: str, low: float, high: float):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: Trial) -> Any:
        return trial.suggest_loguniform(self.name, self.low, self.high)

    def __repr__(self) -> str:
        return f"LogUniformSuggestion(name={self.name!r}, low={self.low!r}, high={self.high!r})"


class IntegerSuggestion(Suggestion):
    def __init__(self, name: str, low: int, high: int, step: int = 1):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high
        self.step = step

    def suggest(self, trial: Trial) -> Any:
        return trial.suggest_int(self.name, self.low, self.high, step=self.step)

    def __repr__(self) -> str:
        return f"IntegerSuggestion(name={self.name!r}, low={self.low!r}, high={self.high!r})"


class IntegerLogUniformSuggestion(Suggestion):
    def __init__(self, name: str, low: int, high: int):
        super().__init__(name)
        if low > high:
            raise ValueError("Got low > high.")

        self.low = low
        self.high = high

    def suggest(self, trial: Trial) -> Any:
        return round(trial.suggest_loguniform(self.name, self.low, self.high))

    def __repr__(self) -> str:
        return f"IntegerLogUniformSuggestion(name={self.name!r}, low={self.low!r}, high={self.high!r})"


class CategoricalSuggestion(Suggestion):
    def __init__(self, name: str, choices: List[Any]):
        super().__init__(name)
        self.choices = choices

    def suggest(self, trial: Trial) -> Any:
        return trial.suggest_categorical(self.name, self.choices)

    def __repr__(self) -> str:
        return (
            f"IntegerLogUniformSuggestion(name={self.name!r}, choices={self.choices!r})"
        )



from typing import Optional

from sklearn.preprocessing import normalize

from irspack.utils import get_n_threads

from ..definitions import InteractionMatrix
from ._knn import RP3betaComputer
from .base import BaseSimilarityRecommender


class RP3betaRecommender(BaseSimilarityRecommender):
    """3-Path random walk with the item-popularity penalization:

        - `Updatable, Accurate, Diverse, and Scalable Recommendations for Interactive Applications
          <https://dl.acm.org/doi/10.1145/2955101>`_

    The version here implements its view as KNN-based method, as pointed out in

        - `A Troubling Analysis of Reproducibility and Progress in Recommender Systems Research
          <https://arxiv.org/abs/1911.07698>`_

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        alpha (float, optional): The power to which ``X_train_all`` is exponentiated.
            Defaults to 1. Note that this has no effect if all the entries in
            ``X_train_all`` are equal.
        beta (float, optional): Specifies how the user->item transition
            probability will be penalized by the popularity. Defaults to 0.6.
        top_k (Optional[int], optional): Maximal number of non-zero entries retained
        for each column of the similarity matrix ``W``.
        normalize_weight (bool, optional): Whether to perform row-wise normalization of ``W``.
            Defaults to False.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to 1. Defaults to None.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        alpha: float = 1,
        beta: float = 0.6,
        top_k: Optional[int] = None,
        normalize_weight: bool = False,
        n_threads: Optional[int] = None,
    ):
        super().__init__(X_train_all)
        self.alpha = alpha
        self.beta = beta
        self.top_k = top_k
        self.normalize_weight = normalize_weight
        self.n_threads = get_n_threads(n_threads)

    def _learn(self) -> None:
        computer = RP3betaComputer(
            self.X_train_all.T,
            alpha=self.alpha,
            beta=self.beta,
            n_threads=self.n_threads,
        )
        top_k = self.X_train_all.shape[1] if self.top_k is None else self.top_k
        self.W_ = computer.compute_W(self.X_train_all.T, top_k)
        if self.normalize_weight:
            self.W_ = normalize(self.W_, norm="l1", axis=1)



from ..utils import rowwise_train_test_split
from .random import UserTrainTestInteractionPair, split_dataframe_partial_user_holdout
from .specified import holdout_specific_interactions

__all__ = [
    "UserTrainTestInteractionPair",
    "split_dataframe_partial_user_holdout",
    "holdout_specific_interactions",
    "rowwise_train_test_split",
]



from irspack.utils import get_n_threads

from ..definitions import InteractionMatrix
from ._rwr import RandomWalkGenerator
from .base import BaseSimilarityRecommender


class RandomWalkWithRestartRecommender(BaseSimilarityRecommender):
    def __init__(
        self,
        X_train_all: InteractionMatrix,
        decay: float = 0.3,
        cutoff: int = 1000,
        n_samples: int = 1000,
        random_seed: int = 42,
        n_threads: int = 4,
    ):
        super().__init__(X_train_all)
        self.decay = decay
        self.n_samples = n_samples
        self.cutoff = cutoff
        self.random_seed = random_seed
        self.n_threads = get_n_threads(n_threads)

    def _learn(self) -> None:
        rwg = RandomWalkGenerator(self.X_train_all.tocsr())
        self.W_ = rwg.run_with_restart(
            self.decay,
            self.cutoff,
            self.n_samples,
            self.n_threads,
            self.random_seed,
        )
        self.W_ = self.W_.tocsc() / self.n_samples



from abc import ABC, abstractmethod
from typing import Any, List

from scipy import sparse as sps
from scipy.sparse.csr import csr_matrix


class BaseEncoder(ABC):
    names: List[Any]

    def __init___(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def transform_sparse(self, arg: List[Any]) -> sps.csr_matrix:
        raise NotImplementedError("not implemented")

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError("not implemented")



"""Backend C++ inplementation for Random walk with restart."""
from typing import Iterable as iterable
from typing import Iterator as iterator
from typing import *

from numpy import float32, float64, int32

_Shape = Tuple[int, ...]
import scipy.sparse

__all__ = ["RandomWalkGenerator"]

class RandomWalkGenerator:
    def __init__(self, arg0: scipy.sparse.csr_matrix[float32]) -> None: ...
    def run_with_restart(
        self, arg0: float, arg1: int, arg2: int, arg3: int, arg4: int
    ) -> scipy.sparse.csr_matrix[int32]: ...
    pass



from typing import Dict

import numpy as np
import scipy.sparse as sps
from sklearn.linear_model import ElasticNet

from irspack.recommenders import SLIMRecommender


def test_slim_positive(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    alpha = 0.1
    l1_ratio = 0.5
    X = test_interaction_data["X_small"]
    rec = SLIMRecommender(
        X, alpha=alpha, l1_ratio=l1_ratio, positive_only=True, n_iter=10, n_threads=8
    )
    rec.learn()

    enet = ElasticNet(
        alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False, positive=True, max_iter=10
    )
    for iind in range(rec.W.shape[1]):
        m = rec.W[:, iind].toarray().ravel()
        Xcp = X.toarray()
        y = X[:, iind].toarray().ravel()
        Xcp[:, iind] = 0.0
        enet.fit(Xcp, y)
        np.testing.assert_allclose(enet.coef_, m, rtol=1e-2)


def test_slim_allow_negative(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    alpha = 0.1
    l1_ratio = 0.5
    X = test_interaction_data["X_small"]
    rec = SLIMRecommender(
        X, alpha=alpha, l1_ratio=l1_ratio, positive_only=False, n_iter=10, n_threads=8
    )
    rec.learn()

    enet = ElasticNet(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False, max_iter=10)
    for iind in range(rec.W.shape[1]):
        m = rec.W[:, iind].toarray().ravel()
        Xcp = X.toarray()
        y = X[:, iind].toarray().ravel()
        Xcp[:, iind] = 0.0
        enet.fit(Xcp, y)
        np.testing.assert_allclose(enet.coef_, m, rtol=1e-2)



import warnings
from typing import List, Type

from ..optimizers.base_optimizer import BaseOptimizer, BaseOptimizerWithEarlyStopping
from ..parameter_tuning import (
    CategoricalSuggestion,
    IntegerSuggestion,
    LogUniformSuggestion,
    Suggestion,
    UniformSuggestion,
)
from ..recommenders import (
    AsymmetricCosineKNNRecommender,
    CosineKNNRecommender,
    DenseSLIMRecommender,
    IALSRecommender,
    JaccardKNNRecommender,
    NMFRecommender,
    P3alphaRecommender,
    RandomWalkWithRestartRecommender,
    RP3betaRecommender,
    SLIMRecommender,
    TopPopRecommender,
    TruncatedSVDRecommender,
    TverskyIndexKNNRecommender,
)

default_tune_range_knn = [
    IntegerSuggestion("top_k", 4, 1000),
    UniformSuggestion("shrinkage", 0, 1000),
]

default_tune_range_knn_with_weighting = [
    IntegerSuggestion("top_k", 4, 1000),
    UniformSuggestion("shrinkage", 0, 1000),
    CategoricalSuggestion("feature_weighting", ["NONE", "TF_IDF", "BM_25"]),
]


_BaseOptimizerArgsString = """Args:
    data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
        The train data.
    val_evaluator (Evaluator):
        The validation evaluator which measures the performance of the recommenders.
    metric (str, optional) :
        Target metric. Defaults to "ndcg".
    logger (Optional[logging.Logger], optional) :
        The logger used during the optimization steps. Defaults to None.
        If ``None``, the default logger of irspack will be used.
    suggest_overwrite (List[Suggestion], optional) :
        Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
        the default parameter search space defined by ``default_tune_range``
        Defaults to list().
    fixed_params (Dict[str, Any], optional):
        Fixed parameters passed to recommenders during the optimization procedure.
        If such a parameter exists in ``default_tune_range``, it will not be tuned.
        Defaults to dict().
"""

_BaseOptimizerWithEarlyStoppingArgsString = """Args:
    data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
        The train data.
    val_evaluator (Evaluator):
        The validation evaluator which measures the performance of the recommenders.
    metric (str, optional): Target metric. Defaults to "ndcg".
    logger (Optional[logging.Logger], optional):
        The logger used during the optimization steps. Defaults to None.
        If ``None``, the default logger of irspack will be used.
    suggest_overwrite (List[Suggestion], optional):
        Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
        the default parameter search space defined by ``default_tune_range``
        Defaults to list().
    fixed_params (Dict[str, Any], optional):
        Fixed parameters passed to recommenders during the optimization procedure.
        If such a parameter exists in ``default_tune_range``, it will not be tuned.
        Defaults to dict().
    max_epoch (int, optional):
        The maximal number of epochs for the training. Defaults to 512.
    validate_epoch (int, optional):
        The frequency of validation score measurement. Defaults to 5.
    score_degradation_max (int, optional):
        Maximal number of allowed score degradation. Defaults to 5. Defaults to 5.
"""


def _add_docstring(
    cls: Type[BaseOptimizer], args: str = _BaseOptimizerArgsString
) -> None:

    if cls.default_tune_range:

        ranges = ""
        for suggest in cls.default_tune_range:
            ranges += f"          - ``{suggest!r}``\n"

        ranges += "\n"
        tune_range = f"""The default tune range is

{ranges}"""
    else:
        tune_range = "   There is no tunable parameters."
    docs = f"""Optimizer class for :class:`irspack.recommenders.{cls.recommender_class.__name__}`.

{tune_range}

{args}

    """
    cls.__doc__ = docs


class TopPopOptimizer(BaseOptimizer):
    default_tune_range: List[Suggestion] = []
    recommender_class = TopPopRecommender


_add_docstring(TopPopOptimizer)


class IALSOptimizer(BaseOptimizerWithEarlyStopping):
    default_tune_range = [
        IntegerSuggestion("n_components", 4, 200),
        LogUniformSuggestion("alpha", 1, 50),
        LogUniformSuggestion("reg", 1e-10, 1e-2),
    ]
    recommender_class = IALSRecommender


_add_docstring(IALSOptimizer, _BaseOptimizerWithEarlyStoppingArgsString)


class P3alphaOptimizer(BaseOptimizer):
    default_tune_range = [
        LogUniformSuggestion("alpha", low=1e-10, high=2),
        IntegerSuggestion("top_k", low=10, high=1000),
        CategoricalSuggestion("normalize_weight", [True, False]),
    ]
    recommender_class = P3alphaRecommender


_add_docstring(P3alphaOptimizer)


class DenseSLIMOptimizer(BaseOptimizer):
    default_tune_range = [LogUniformSuggestion("reg", 1, 1e4)]
    recommender_class = DenseSLIMRecommender


_add_docstring(DenseSLIMOptimizer)


class RP3betaOptimizer(BaseOptimizer):
    default_tune_range = [
        LogUniformSuggestion("alpha", 1e-5, 10),
        IntegerSuggestion("top_k", 2, 1000),
        LogUniformSuggestion("beta", 1e-5, 5e-1),
        CategoricalSuggestion("normalize_weight", [True, False]),
    ]
    recommender_class = RP3betaRecommender


_add_docstring(RP3betaOptimizer)


class TruncatedSVDOptimizer(BaseOptimizer):
    default_tune_range = [IntegerSuggestion("n_components", 4, 512)]
    recommender_class = TruncatedSVDRecommender


_add_docstring(TruncatedSVDOptimizer)


class RandomWalkWithRestartOptimizer(BaseOptimizer):
    default_tune_range = [
        UniformSuggestion("decay", 1e-2, 9.9e-1),
        IntegerSuggestion("n_samples", 100, 2000, step=100),
        IntegerSuggestion("cutoff", 100, 2000, step=100),
    ]

    recommender_class = RandomWalkWithRestartRecommender


_add_docstring(RandomWalkWithRestartOptimizer)


class SLIMOptimizer(BaseOptimizer):
    default_tune_range = [
        LogUniformSuggestion("alpha", 1e-5, 1),
        UniformSuggestion("l1_ratio", 0, 1),
    ]
    recommender_class = SLIMRecommender


_add_docstring(SLIMOptimizer)


class NMFOptimizer(BaseOptimizer):
    default_tune_range = [
        IntegerSuggestion("n_components", 4, 512),
        LogUniformSuggestion("alpha", 1e-10, 1e-1),
        UniformSuggestion("l1_ratio", 0, 1),
        CategoricalSuggestion("beta_loss", ["frobenius", "kullback-leibler"]),
    ]
    recommender_class = NMFRecommender


_add_docstring(NMFOptimizer)


class CosineKNNOptimizer(BaseOptimizer):
    default_tune_range = default_tune_range_knn_with_weighting.copy() + [
        CategoricalSuggestion("normalize", [False, True])
    ]

    recommender_class = CosineKNNRecommender


_add_docstring(CosineKNNOptimizer)


class JaccardKNNOptimizer(BaseOptimizer):

    default_tune_range = default_tune_range_knn.copy()
    recommender_class = JaccardKNNRecommender


_add_docstring(JaccardKNNOptimizer)


class TverskyIndexKNNOptimizer(BaseOptimizer):
    default_tune_range = default_tune_range_knn.copy() + [
        UniformSuggestion("alpha", 0, 2),
        UniformSuggestion("beta", 0, 2),
    ]

    recommender_class = TverskyIndexKNNRecommender


_add_docstring(TverskyIndexKNNOptimizer)


class AsymmetricCosineKNNOptimizer(BaseOptimizer):
    default_tune_range = default_tune_range_knn_with_weighting + [
        UniformSuggestion("alpha", 0, 1)
    ]

    recommender_class = AsymmetricCosineKNNRecommender


_add_docstring(AsymmetricCosineKNNOptimizer)

try:
    from ..recommenders.bpr import BPRFMRecommender

    class BPRFMOptimizer(BaseOptimizerWithEarlyStopping):
        default_tune_range = [
            IntegerSuggestion("n_components", 4, 256),
            LogUniformSuggestion("item_alpha", 1e-9, 1e-2),
            LogUniformSuggestion("user_alpha", 1e-9, 1e-2),
            CategoricalSuggestion("loss", ["bpr", "warp"]),
        ]
        recommender_class = BPRFMRecommender

    _add_docstring(BPRFMOptimizer, _BaseOptimizerWithEarlyStoppingArgsString)


except:
    pass


try:
    from ..recommenders.multvae import MultVAERecommender

    class MultVAEOptimizer(BaseOptimizerWithEarlyStopping):
        default_tune_range = [
            CategoricalSuggestion("dim_z", [32, 64, 128, 256]),
            CategoricalSuggestion("enc_hidden_dims", [128, 256, 512]),
            CategoricalSuggestion("kl_anneal_goal", [0.1, 0.2, 0.4]),
        ]
        recommender_class = MultVAERecommender

    _add_docstring(MultVAEOptimizer, _BaseOptimizerWithEarlyStoppingArgsString)


except:
    warnings.warn("MultVAEOptimizer is not available.")
    pass



from numpy import float32
import irspack.utils._util_cpp
from typing import *
from typing import Iterable as iterable
from typing import Iterator as iterator
from numpy import float64
_Shape = Tuple[int, ...]
import scipy.sparse
__all__  = [
"okapi_BM_25_weight",
"remove_diagonal",
"rowwise_train_test_split_d",
"rowwise_train_test_split_f",
"rowwise_train_test_split_i",
"slim_weight_allow_negative",
"slim_weight_positive_only",
"sparse_mm_threaded",
"tf_idf_weight"
]
def okapi_BM_25_weight(X: scipy.sparse.csr_matrix[float64], k1: float = 1.2, b: float = 0.75) -> scipy.sparse.csr_matrix[float64]:
    pass
def remove_diagonal(arg0: scipy.sparse.csr_matrix[float64]) -> scipy.sparse.csr_matrix[float64]:
    pass
def rowwise_train_test_split_d(arg0: scipy.sparse.csr_matrix[float64], arg1: float, arg2: int) -> Tuple[scipy.sparse.csr_matrix[float64], scipy.sparse.csr_matrix[float64]]:
    pass
def rowwise_train_test_split_f(arg0: scipy.sparse.csr_matrix[float32], arg1: float, arg2: int) -> Tuple[scipy.sparse.csr_matrix[float32], scipy.sparse.csr_matrix[float32]]:
    pass
def rowwise_train_test_split_i(arg0: scipy.sparse.csr_matrix[float32], arg1: float, arg2: int) -> Tuple[scipy.sparse.csr_matrix[float32], scipy.sparse.csr_matrix[float32]]:
    pass
def slim_weight_allow_negative(X: scipy.sparse.csr_matrix[float32], n_threads: int, n_iter: int, l2_coeff: float, l1_coeff: float) -> scipy.sparse.csc_matrix[float32]:
    pass
def slim_weight_positive_only(X: scipy.sparse.csr_matrix[float32], n_threads: int, n_iter: int, l2_coeff: float, l1_coeff: float) -> scipy.sparse.csc_matrix[float32]:
    pass
def sparse_mm_threaded(arg0: scipy.sparse.csr_matrix[float64], arg1: scipy.sparse.csc_matrix[float64], arg2: int) -> scipy.sparse.csr_matrix[float64]:
    pass
def tf_idf_weight(X: scipy.sparse.csr_matrix[float64], smooth: bool = True) -> scipy.sparse.csr_matrix[float64]:
    pass



m: int
n: int
from typing import Iterable as iterable
from typing import Iterator as iterator
from typing import *

from numpy import float32, float64

import irspack.recommenders._ials

_Shape = Tuple[int, ...]
import numpy
import scipy.sparse

__all__ = ["IALSLearningConfig", "IALSLearningConfigBuilder", "IALSTrainer"]

class IALSLearningConfig:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self,
        arg0: int,
        arg1: float,
        arg2: float,
        arg3: float,
        arg4: int,
        arg5: int,
        arg6: bool,
        arg7: int,
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    pass

class IALSLearningConfigBuilder:
    def __init__(self) -> None: ...
    def build(self) -> IALSLearningConfig: ...
    def set_K(self, arg0: int) -> IALSLearningConfigBuilder: ...
    def set_alpha(self, arg0: float) -> IALSLearningConfigBuilder: ...
    def set_init_stdev(self, arg0: float) -> IALSLearningConfigBuilder: ...
    def set_max_cg_steps(self, arg0: int) -> IALSLearningConfigBuilder: ...
    def set_n_threads(self, arg0: int) -> IALSLearningConfigBuilder: ...
    def set_random_seed(self, arg0: int) -> IALSLearningConfigBuilder: ...
    def set_reg(self, arg0: float) -> IALSLearningConfigBuilder: ...
    def set_use_cg(self, arg0: bool) -> IALSLearningConfigBuilder: ...
    pass

class IALSTrainer:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self, arg0: IALSLearningConfig, arg1: scipy.sparse.csr_matrix[float32]
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def step(self) -> None: ...
    def transform_item(
        self, arg0: scipy.sparse.csr_matrix[float32]
    ) -> numpy.ndarray[float32, _Shape[m, n]]: ...
    def transform_user(
        self, arg0: scipy.sparse.csr_matrix[float32]
    ) -> numpy.ndarray[float32, _Shape[m, n]]: ...
    def user_scores(
        self, arg0: int, arg1: int
    ) -> numpy.ndarray[float32, _Shape[m, n]]: ...
    @property
    def item(self) -> numpy.ndarray[float32, _Shape[m, n]]:
        """
        :type: numpy.ndarray[float32, _Shape[m, n]]
        """
    @item.setter
    def item(self, arg0: numpy.ndarray[float32, _Shape[m, n]]) -> None:
        pass
    @property
    def user(self) -> numpy.ndarray[float32, _Shape[m, n]]:
        """
        :type: numpy.ndarray[float32, _Shape[m, n]]
        """
    @user.setter
    def user(self, arg0: numpy.ndarray[float32, _Shape[m, n]]) -> None:
        pass
    pass



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.recommenders import TruncatedSVDRecommender

RNS = np.random.RandomState(0)

X = RNS.rand(200, 512)
X[X <= 0.9] = 0
X[X > 0.9] = 1
X = sps.csr_matrix(X)


def test_truncsvd() -> None:
    with pytest.warns(UserWarning):
        overfit_rec = TruncatedSVDRecommender(X, n_components=X.shape[1] + 1)

    with pytest.raises(RuntimeError):
        overfit_rec.get_score_block(0, X.shape[0])

    with pytest.raises(RuntimeError):
        overfit_rec.get_item_embedding()

    with pytest.raises(RuntimeError):
        overfit_rec.get_user_embedding()

    overfit_rec.learn()
    X_reproduced = overfit_rec.get_score_block(0, X.shape[0])
    X_frobenius = (X.data ** 2).sum()
    residual = (X_reproduced - X).A1
    assert ((residual ** 2).sum() / X_frobenius) < 1e-3

    X_reproduced_cold = overfit_rec.get_score_cold_user(X)
    residual_cold = (X_reproduced_cold - X).A1
    assert ((residual_cold ** 2).sum() / X_frobenius) < 1e-3



from typing import List

import numpy as np
from scipy import sparse as sps

from .base import BaseEncoder


class BinningEncoder(BaseEncoder):
    def __init__(self, items: List[float], n_percentiles: int = 30):
        super().__init__()
        ps = np.linspace(0, 100, n_percentiles)
        items_farray: np.ndarray = np.asfarray(items)
        self.percentiles = np.unique(
            np.percentile(items_farray[~np.isnan(items_farray)], ps)
        )
        self.names = ["NaN"]
        self.names.append(f"(-infty, {self.percentiles[0]}]")
        for i in range(1, self.percentiles.shape[0]):
            self.names.append(f"({self.percentiles[i-1]} , {self.percentiles[i]}]")
        self.names.append(f"({self.percentiles[-1]}, infty)")

    def transform_sparse(self, items: List[str]) -> sps.csr_matrix:
        """Transform sparse"""
        items_farray: np.ndarray = np.asarray(items)
        rows = np.arange(items_farray.shape[0])
        cols = np.zeros(items_farray.shape[0], dtype=np.int32)
        valid_mask = np.where(~np.isnan(items_farray))[0]
        valid_items = items_farray[valid_mask]
        cols[valid_mask] += 1
        for v in self.percentiles:
            cols[valid_mask] += valid_items > v
        data = np.ones(rows.shape[0], dtype=np.float64)
        result = sps.csr_matrix((data, (rows, cols)), shape=(rows.shape[0], len(self)))
        assert result.shape[1] == len(self)
        return result

    def __len__(self) -> int:
        return len(self.percentiles) + 2



import pickle
from logging import getLogger
from typing import IO, Any

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from irspack.evaluator import Evaluator
from irspack.optimizers.base_optimizer import BaseOptimizerWithEarlyStopping
from irspack.recommenders.base_earlystop import (
    BaseRecommenderWithEarlyStopping,
    TrainerBase,
)
from irspack.split import rowwise_train_test_split

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(100, 32) > 0.5).astype(np.float64)
)
X_large = sps.csr_matrix(
    (np.random.RandomState(42).rand(50, 200) > 0.5).astype(np.float64)
)


class MockTrainer(TrainerBase):
    def __init__(self) -> None:
        self.epoch: int = 0

    def load_state(self, ifs: IO) -> None:
        self.epoch = pickle.load(ifs)["epoch"]

    def save_state(self, ofs: IO) -> None:
        pickle.dump(dict(epoch=self.epoch), ofs)

    def run_epoch(self) -> None:
        self.epoch += 1


class MockRecommender(BaseRecommenderWithEarlyStopping):
    trainer_class = MockTrainer

    def __init__(
        self,
        X: InteractionMatrix,
        X_test: InteractionMatrix,
        target_epoch: int = 20,
        **kwargs: Any
    ):
        super().__init__(X, **kwargs)
        self.answer = np.asfarray(X_test)
        self.target_epoch = target_epoch
        self.rns = np.random.RandomState(42)

    def _create_trainer(self) -> MockTrainer:
        return MockTrainer()

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        assert isinstance(self.trainer, MockTrainer)
        score = self.answer[user_indices]
        score = (
            score
            + 10
            * self.rns.randn(*score.shape)
            * abs(self.target_epoch - self.trainer.epoch)
            / self.target_epoch
        )
        return score


class MockOptimizer(BaseOptimizerWithEarlyStopping):
    recommender_class = MockRecommender


@pytest.mark.parametrize(
    "X, target_epoch", [(X_small, 20), (X_small, 15), (X_large, 5)]
)
def test_optimizer_by_mock(X: InteractionMatrix, target_epoch: int) -> None:
    X_train, X_val = rowwise_train_test_split(X)
    evaluator = Evaluator(X_val, 0)
    optimizer = MockOptimizer(
        X_train,
        evaluator,
        fixed_params=dict(X_test=X_val.toarray(), target_epoch=target_epoch),
        logger=getLogger("IGNORE"),
    )
    config, _ = optimizer.optimize(n_trials=1, random_seed=42)
    assert config["max_epoch"] == target_epoch



"""
An example for user cold-start scenario.
CB2CF Requires the following additional dependencies to run.

 - lightfm
 - jax
 - jaxlib
 - dm-haiku
 - optax

"""
import json
import os
from typing import Any, Dict, List, Tuple, Type

import numpy as np
import pandas as pd
from scipy import sparse as sps
from sklearn.model_selection import train_test_split

from irspack.dataset.movielens import MovieLens1MDataManager
from irspack.definitions import UserIndexArray
from irspack.user_cold_start import (
    BaseUserColdStartOptimizer,
    CB2BPRFMOptimizer,
    CB2IALSOptimizer,
    CB2TruncatedSVDOptimizer,
    LinearMethodOptimizer,
    TopPopularOptimizer,
    UserCBCosineKNNOptimizer,
    UserColdStartEvaluator,
)
from irspack.utils.encoders import (
    BinningEncoder,
    CategoricalValueEncoder,
    DataFrameEncoder,
)

os.environ["OMP_NUM_THREADS"] = "8"
os.environ["IRSPACK_NUM_THREADS_DEFAULT"] = "8"


if __name__ == "__main__":
    BASE_CUTOFF = 20
    loader = MovieLens1MDataManager()
    user_df = loader.read_user_info()
    ratings = loader.read_interaction()
    user_ids_unique: List[int] = np.unique(ratings.userId)
    movie_ids_unique: List[int] = np.unique(ratings.movieId)
    movie_id_to_index = {id_: i for i, id_ in enumerate(movie_ids_unique)}

    def df_to_sparse(df: pd.DataFrame) -> Tuple[UserIndexArray, sps.csr_matrix]:
        unique_uids, row = np.unique(df.userId, return_inverse=True)
        col = df.movieId.map(movie_id_to_index)
        return (
            unique_uids,
            sps.csr_matrix(
                (np.ones(df.shape[0]), (row, col)),
                shape=(len(unique_uids), len(movie_id_to_index)),
            ),
        )

    train_uids, test_uids = train_test_split(
        user_ids_unique, test_size=0.2, random_state=42
    )
    train_uids, X_train = df_to_sparse(ratings[ratings.userId.isin(train_uids)])
    test_uids, X_test = df_to_sparse(ratings[ratings.userId.isin(test_uids)])

    ### Preprocess user data
    user_df["zip_first"] = user_df.zipcode.str[0]
    columns = ["occupation", "zip_first"]
    encoder_all = DataFrameEncoder()
    encoder_all.add_column(
        "gender", CategoricalValueEncoder(user_df["gender"])
    ).add_column("age", BinningEncoder(user_df["age"], n_percentiles=10)).add_column(
        "occupation", CategoricalValueEncoder(user_df["occupation"])
    ).add_column(
        "zip_first", CategoricalValueEncoder(user_df["zip_first"])
    )

    user_info_train = encoder_all.transform_sparse(user_df.reindex(train_uids))
    user_info_test = encoder_all.transform_sparse(user_df.reindex(test_uids))

    test_evaluator = UserColdStartEvaluator(X_test, user_info_test)

    trial_configurations: List[Tuple[Type[BaseUserColdStartOptimizer], int]] = [
        (TopPopularOptimizer, 1),
        (UserCBCosineKNNOptimizer, 20),
        (LinearMethodOptimizer, 10),
    ]
    test_results: List[Dict[str, Any]] = []
    for optimizer_class, n_trials in trial_configurations:
        best_param = optimizer_class.split_and_optimize(
            X_train, user_info_train, n_trials=n_trials
        )
        rec = optimizer_class.recommender_class(X_train, user_info_train, **best_param)
        rec.learn()
        test_results.append(
            dict(
                recommender_name=optimizer_class.recommender_class.__name__,
                test_result=test_evaluator.get_score(rec),
            )
        )
    for cboptim_class in [
        CB2TruncatedSVDOptimizer,
        CB2IALSOptimizer,
        CB2BPRFMOptimizer,
    ]:
        rec, best_config, best_nn_config = cboptim_class.split_and_optimize(
            X_train, user_info_train, n_trials=20
        )
        test_results.append(
            dict(
                recommender_name=cboptim_class.recommender_class.__name__,
                test_result=test_evaluator.get_score(rec),
            )
        )
    with open("test_results.json", "w") as ofs:
        json.dump(test_results, ofs, indent=2)



import warnings
from typing import Optional

from numpy import random
from sklearn.decomposition import TruncatedSVD

from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)
from .base import BaseRecommenderWithItemEmbedding, BaseRecommenderWithUserEmbedding


class TruncatedSVDRecommender(
    BaseRecommenderWithUserEmbedding,
    BaseRecommenderWithItemEmbedding,
):
    """Use (randomized) SVD to factorize the input matrix into low-rank matrices.

    Args:
        X_train_all:
            Input interaction matrix.

        n_components:
            The rank of approximation. Defaults to 4.
            If this is larger than X_train_all, the value will be truncated into ``X_train_all.shape[1]``

        random_seed:
            The random seed to be passed on core TruncSVD.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        n_components: int = 4,
        random_seed: int = 0,
    ) -> None:

        assert X_train_all.shape[1] > 1
        super().__init__(X_train_all)
        if n_components >= self.X_train_all.shape[1]:
            warnings.warn(
                "n_components >= than X_train_all.shape[1]. Set it to X_train_all.shape[1] - 1."
            )
            n_components = self.X_train_all.shape[1] - 1
        self.n_components = n_components
        self.decomposer_: Optional[TruncatedSVD] = None
        self.z_: Optional[DenseMatrix] = None
        self.random_seed = random_seed

    @property
    def z(self) -> DenseMatrix:
        if self.z_ is None:
            raise RuntimeError("z fetched before fit")
        return self.z_

    @property
    def decomposer(self) -> TruncatedSVD:
        if self.decomposer_ is None:
            raise RuntimeError("decomposer fetched before fit.")
        return self.decomposer_

    def _learn(self) -> None:
        self.decomposer_ = TruncatedSVD(
            n_components=self.n_components, random_state=self.random_seed
        )
        self.z_ = self.decomposer_.fit_transform(self.X_train_all)

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.z[user_indices].dot(self.decomposer.components_)

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return self.z[begin:end].dot(self.decomposer.components_)

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        return self.decomposer.transform(X).dot(self.decomposer.components_)

    def get_user_embedding(self) -> DenseMatrix:
        return self.z

    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return user_embedding.dot(self.decomposer.components_)

    def get_item_embedding(self) -> DenseMatrix:
        return self.decomposer.components_.T

    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return self.z[user_indices].dot(item_embedding.T)



from logging import Logger
from typing import Any, Dict, List, Optional, Tuple, Type

import numpy as np
from scipy import sparse as sps

from irspack.evaluator import Evaluator as Evaluator_hot
from irspack.optimizers import IALSOptimizer, TruncatedSVDOptimizer
from irspack.optimizers.base_optimizer import BaseOptimizer as HotBaseOptimizer
from irspack.parameter_tuning import Suggestion
from irspack.recommenders import IALSRecommender, TruncatedSVDRecommender
from irspack.user_cold_start.recommenders.base import BaseUserColdStartRecommender
from irspack.utils.default_logger import get_default_logger

from ..definitions import DenseMatrix, DenseScoreArray, InteractionMatrix, ProfileMatrix
from ..recommenders.base import BaseRecommenderWithUserEmbedding
from ..utils import rowwise_train_test_split
from ..utils.nn import MLP, MLPOptimizer, MLPSearchConfig, MLPTrainingConfig


class CB2CFUserColdStartRecommender(BaseUserColdStartRecommender):
    def __init__(self, cf_rec: BaseRecommenderWithUserEmbedding, mlp: MLP):
        self.cf_rec = cf_rec
        self.mlp = mlp

    def _learn(self) -> None:
        pass

    def get_score(self, profile: sps.csr_matrix) -> DenseScoreArray:
        user_embedding: DenseMatrix = self.mlp.predict(
            profile.astype(np.float32).toarray()
        )
        return self.cf_rec.get_score_from_user_embedding(user_embedding).astype(
            np.float64
        )


class CB2CFUserOptimizerBase(object):
    recommender_class: Type[BaseRecommenderWithUserEmbedding]
    cf_optimizer_class: Type[HotBaseOptimizer]

    def __init__(
        self,
        X_cf_train_all: InteractionMatrix,
        hot_evaluator: Evaluator_hot,
        X_profile: ProfileMatrix,
        nn_search_config: Optional[MLPSearchConfig] = None,
    ):
        assert X_cf_train_all.shape[0] == X_profile.shape[0]

        X_val_ground_truth = hot_evaluator.core.get_ground_truth()
        assert X_val_ground_truth.shape[1] == X_cf_train_all.shape[1]
        assert (
            hot_evaluator.offset + X_val_ground_truth.shape[0]
            <= X_cf_train_all.shape[0]
        )
        self.X_profile = X_profile
        self.X_cf_train_all = X_cf_train_all
        self.hot_evaluator = hot_evaluator

        # reconstruct all the available interaction
        X_all = X_cf_train_all + sps.vstack(
            [
                sps.csr_matrix(
                    (hot_evaluator.offset, X_cf_train_all.shape[1]),
                    dtype=np.float64,
                ),
                X_val_ground_truth,
                sps.csr_matrix(
                    (
                        X_cf_train_all.shape[0]
                        - hot_evaluator.offset
                        - X_val_ground_truth.shape[0],
                        X_cf_train_all.shape[1],
                    ),
                    dtype=np.float64,
                ),
            ],
            format="csr",
        )
        self.X_all = X_all

        if nn_search_config is None:
            nn_search_config = MLPSearchConfig()
        self.nn_search_config = nn_search_config

    @classmethod
    def split_and_optimize(
        cls,
        X_all: InteractionMatrix,
        X_profile: ProfileMatrix,
        cf_evaluator_config: Dict[str, Any] = dict(
            cutoff=20, n_threads=4, target_metric="ndcg"
        ),
        cf_split_config: Dict[str, Any] = dict(random_seed=42, test_ratio=0.2),
        nn_search_config: Optional[MLPSearchConfig] = None,
        n_trials: int = 20,
    ) -> Tuple[CB2CFUserColdStartRecommender, Dict[str, Any], MLPTrainingConfig]:
        assert X_all.shape[0] == X_profile.shape[0]
        X_all = X_all
        X_profile = X_profile
        X_tr_, X_val_ = rowwise_train_test_split(X_all, **cf_split_config)
        hot_evaluator = Evaluator_hot(X_val_, 0, **cf_evaluator_config)
        optimizer = cls(
            X_tr_,
            hot_evaluator,
            X_profile,
            nn_search_config=nn_search_config,
        )
        return optimizer.search_all(n_trials=n_trials)

    def search_all(
        self,
        n_trials: int = 40,
        logger: Optional[Logger] = None,
        timeout: Optional[int] = None,
        reconstruction_search_config: Optional[MLPSearchConfig] = None,
        cf_suggest_overwrite: List[Suggestion] = [],
        cf_fixed_params: Dict[str, Any] = dict(),
        random_seed: Optional[int] = None,
    ) -> Tuple[CB2CFUserColdStartRecommender, Dict[str, Any], MLPTrainingConfig]:
        if logger is None:
            logger = get_default_logger()
        recommender, best_config_recommender = self.search_embedding(
            n_trials,
            logger,
            timeout=timeout,
            suggest_overwrite=cf_suggest_overwrite,
            fixed_params=cf_fixed_params,
            random_seed=random_seed,
        )

        if logger is not None:
            logger.info("Start learning feature -> embedding map.")

        mlp, best_config_mlp = self.search_reconstruction(
            recommender,
            n_trials,
            logger=logger,
            config=reconstruction_search_config,
            random_seed=random_seed,
        )

        return (
            CB2CFUserColdStartRecommender(recommender, mlp),
            best_config_recommender,
            best_config_mlp,
        )

    def search_embedding(
        self,
        n_trials: int,
        logger: Optional[Logger] = None,
        timeout: Optional[int] = None,
        suggest_overwrite: List[Suggestion] = [],
        fixed_params: Dict[str, Any] = dict(),
        random_seed: Optional[int] = None,
    ) -> Tuple[BaseRecommenderWithUserEmbedding, Dict[str, Any]]:
        searcher = self.cf_optimizer_class(
            self.X_cf_train_all,
            self.hot_evaluator,
            metric="ndcg",
            logger=logger,
            suggest_overwrite=suggest_overwrite,
            fixed_params=fixed_params,
        )
        best_params, _ = searcher.optimize(
            n_trials=n_trials, timeout=timeout, random_seed=random_seed
        )
        rec = self.recommender_class(self.X_all, **best_params)
        rec.learn()
        return rec, best_params

    def search_reconstruction(
        self,
        rec: BaseRecommenderWithUserEmbedding,
        n_trials: int,
        logger: Optional[Logger] = None,
        config: Optional[MLPSearchConfig] = None,
        random_seed: Optional[int] = None,
    ) -> Tuple[MLP, MLPTrainingConfig]:
        embedding = rec.get_user_embedding()
        searcher = MLPOptimizer(self.X_profile, embedding, search_config=config)
        return searcher.search_param_fit_all(
            n_trials=n_trials, logger=logger, random_seed=random_seed
        )


class CB2TruncatedSVDOptimizer(CB2CFUserOptimizerBase):
    recommender_class = TruncatedSVDRecommender
    cf_optimizer_class = TruncatedSVDOptimizer


class CB2IALSOptimizer(CB2CFUserOptimizerBase):
    recommender_class = IALSRecommender
    cf_optimizer_class = IALSOptimizer


__all__ = [
    "CB2CFUserOptimizerBase",
    "CB2IALSOptimizer",
    "CB2TruncatedSVDOptimizer",
]

try:
    from irspack.optimizers import BPRFMOptimizer
    from irspack.recommenders.bpr import BPRFMRecommender

    class CB2BPRFMOptimizer(CB2CFUserOptimizerBase):
        recommender_class = BPRFMRecommender
        cf_optimizer_class = BPRFMOptimizer

    __all__.append("CB2BPRFMOptimizer")


except:
    pass



import os
import re
from typing import List, Tuple

import pandas as pd

from .base import BaseMovieLenstDataLoader


class MovieLens100KDataManager(BaseMovieLenstDataLoader):
    DOWNLOAD_URL = "http://files.grouplens.org/datasets/movielens/ml-100k.zip"
    DEFAULT_PATH = os.path.expanduser("~/.ml-100k.zip")
    INTERACTION_PATH = "ml-100k/u.data"

    USER_INFO_PATH = "ml-100k/u.user"
    ITEM_INFO_PATH = "ml-100k/u.item"
    GENRE_PATH = "ml-100k/u.genre"

    def read_interaction(self) -> pd.DataFrame:
        with self._read_as_istream(self.INTERACTION_PATH) as ifs:
            data = pd.read_csv(
                ifs,
                sep="\t",
                header=None,
                names=["userId", "movieId", "rating", "timestamp"],
            )
            data["timestamp"] = pd.to_datetime(data["timestamp"], unit="s")
            return data

    def read_user_info(self) -> pd.DataFrame:
        with self._read_as_istream(self.USER_INFO_PATH) as ifs:
            return pd.read_csv(
                ifs,
                sep="|",
                header=None,
                names=["userId", "age", "gender", "occupation", "zipcode"],
            ).set_index("userId")

    def _read_genre(self) -> List[str]:
        with self._read_as_istream(self.GENRE_PATH) as ifs:
            items = ifs.read().decode("latin-1").split()
            return [re.sub("\|\d+$", "", i.strip()) for i in items]

    def read_item_info(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        with self._read_as_istream(self.ITEM_INFO_PATH) as ifs:
            df = pd.read_csv(ifs, sep="|", header=None, encoding="latin-1")

        genres = self._read_genre()
        df.columns = [
            "movieId",
            "title",
            "release_date",
            "video_release_date",
            "URL",
        ] + genres
        movie_ids = df.movieId.values
        df["release_date"] = pd.to_datetime(df.release_date)
        genre_df = pd.DataFrame(
            [
                dict(movieId=movie_ids[row], genre=genres[col])
                for row, col in zip(*df[genres].values.nonzero())
            ]
        )
        df = df.set_index("movieId")
        return df.drop(columns=genres), genre_df



import pickle
from typing import IO, Optional

import numpy as np
from lightfm import LightFM

from irspack.utils import get_n_threads

from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)
from .base import BaseRecommenderWithItemEmbedding, BaseRecommenderWithUserEmbedding
from .base_earlystop import BaseRecommenderWithEarlyStopping, TrainerBase


class BPRFMTrainer(TrainerBase):
    def __init__(
        self,
        X: InteractionMatrix,
        n_components: int,
        item_alpha: float,
        user_alpha: float,
        loss: str,
        n_threads: int,
    ):
        self.X = X.tocoo()
        self.n_threads = n_threads
        self.fm = LightFM(
            no_components=n_components,
            item_alpha=item_alpha,
            user_alpha=user_alpha,
            loss=loss,
        )

    def run_epoch(self) -> None:
        self.fm.fit_partial(self.X, num_threads=self.n_threads)

    def load_state(self, ifs: IO) -> None:
        self.fm = pickle.load(ifs)

    def save_state(self, ofs: IO) -> None:
        pickle.dump(self.fm, ofs, protocol=pickle.HIGHEST_PROTOCOL)


class BPRFMRecommender(
    BaseRecommenderWithEarlyStopping,
    BaseRecommenderWithUserEmbedding,
    BaseRecommenderWithItemEmbedding,
):
    """A `LightFM <https://github.com/lyst/lightfm>`_ wrapper for our interface.

    This will create ``LightFM`` instance by

    .. code-block:: python

        fm = LightFM(
            no_components=n_components,
            item_alpha=item_alpha,
            user_alpha=user_alpha,
            loss=loss,
        )

    and run ``fm.fit_partial(X, num_threads=self.n_threads)`` to train through a single epoch.


    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        n_components (int, optional):
            The dimension for latent factor. Defaults to 128.
        item_alpha (float, optional):
            The regularization coefficient for item factors. Defaults to 1e-9.
        user_alpha (float, optional):
            The regularization coefficient for user factors. Defaults to 1e-9.
        loss (str, optional):
            Specifies the loss function type of LightFM. Must be one of {"bpr", "warp"}. Defaults to "bpr".
        validate_epoch (int, optional):
            Frequency of validation score measurement (if any). Defaults to 5.
        score_degradation_max (int, optional):
            Maximal number of allowed score degradation. Defaults to 5.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to 1. Defaults to None.
        max_epoch (int, optional):
            Maximal number of epochs. Defaults to 512.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        n_components: int = 128,
        item_alpha: float = 1e-9,
        user_alpha: float = 1e-9,
        loss: str = "bpr",
        validate_epoch: int = 5,
        score_degradation_max: int = 3,
        n_threads: Optional[int] = None,
        max_epoch: int = 512,
    ):
        super().__init__(
            X_train_all,
            max_epoch=max_epoch,
            validate_epoch=validate_epoch,
            score_degradation_max=score_degradation_max,
        )
        self.n_components = n_components
        self.item_alpha = item_alpha
        self.user_alpha = user_alpha
        if loss not in {"bpr", "warp"}:
            raise ValueError('BPRFM loss must be either "bpr" or "warp".')
        self.loss = loss
        self.trainer: Optional[BPRFMTrainer] = None
        self.n_threads = get_n_threads(n_threads)

    def _create_trainer(self) -> BPRFMTrainer:
        return BPRFMTrainer(
            self.X_train_all,
            self.n_components,
            self.item_alpha,
            self.user_alpha,
            self.loss,
            self.n_threads,
        )

    @property
    def fm(self) -> LightFM:
        if self.trainer is None:
            raise RuntimeError("tried to fetch fm instance before the fit.")
        return self.trainer.fm

    def get_score(self, index: UserIndexArray) -> np.ndarray:
        return (
            self.fm.user_embeddings[index].dot(self.fm.item_embeddings.T)
            + self.fm.item_biases[np.newaxis, :]
        )

    def get_score_block(self, begin: int, end: int) -> np.ndarray:
        return (
            self.fm.user_embeddings[begin:end].dot(self.fm.item_embeddings.T)
            + self.fm.item_biases[np.newaxis, :]
        )

    def get_user_embedding(self) -> DenseMatrix:
        return self.fm.user_embeddings.astype(np.float64)

    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        return (
            user_embedding.dot(self.fm.item_embeddings.T)
            + self.fm.item_biases[np.newaxis, :]
        )

    def get_item_embedding(self) -> DenseMatrix:
        return self.fm.item_embeddings.astype(np.float64)

    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        # ignore bias
        return (
            self.fm.user_embeddings[user_indices]
            .dot(item_embedding.T)
            .astype(np.float64)
        )



import os
import random
from typing import Optional, Tuple

import numpy as np

from ..definitions import InteractionMatrix
from ._util_cpp import (
    okapi_BM_25_weight,
    remove_diagonal,
    rowwise_train_test_split_d,
    rowwise_train_test_split_f,
    rowwise_train_test_split_i,
    sparse_mm_threaded,
    tf_idf_weight,
)


def rowwise_train_test_split(
    X: InteractionMatrix,
    test_ratio: float = 0.5,
    random_seed: Optional[int] = None,
) -> Tuple[InteractionMatrix, InteractionMatrix]:
    """Splits the non-zero elements of a sparse matrix into two (train & test interactions).
    For each row, the ratio of non-zero elements that become the test interaction
    is (approximately) constant.

    Args:
        X:
            The source sparse matrix.
        test_ratio:
            The ratio of test interactions for each row.
            That is, for each row, if it contains ``NNZ``-nonzero elements,
            the number of elements entering into the test interaction
            will be ``math.floor(test_ratio * NNZ)``.
            Defaults to 0.5.
        random_seed:
            The random seed. Defaults to None.

    Returns:
        A tuple of train & test interactions, which sum back to the original matrix.
    """
    if random_seed is None:
        random_seed = random.randint(-(2 ** 32), 2 ** 32 - 1)
    if X.dtype == np.float32:
        return rowwise_train_test_split_f(X, test_ratio, random_seed)
    elif X.dtype == np.float64:
        return rowwise_train_test_split_d(X, test_ratio, random_seed)
    elif X.dtype == np.int64:
        return rowwise_train_test_split_i(X, test_ratio, random_seed)
    else:
        original_dtype = X.dtype
        X_double = X.astype(np.float64)
        X_train_double, X_test_double = rowwise_train_test_split_d(
            X_double, test_ratio, random_seed
        )
        return (
            X_train_double.astype(original_dtype),
            X_test_double.astype(original_dtype),
        )


def get_n_threads(n_threads: Optional[int]) -> int:
    if n_threads is not None:
        return n_threads
    else:
        try:
            n_threads = int(os.environ.get("IRSPACK_NUM_THREADS_DEFAULT", "1"))
            return n_threads
        except:
            raise ValueError(
                'failed to interpret "IRSPACK_NUM_THREADS_DEFAULT" as an integer.'
            )


__all__ = [
    "rowwise_train_test_split",
    "sparse_mm_threaded",
    "okapi_BM_25_weight",
    "tf_idf_weight",
    "remove_diagonal",
]



from collections import Counter
from typing import Any, Generic, List, TypeVar, Union

import numpy as np
from scipy import sparse as sps

from .base import BaseEncoder

T = TypeVar("T")


class CategoricalValueEncoder(Generic[T], BaseEncoder):
    def __init__(self, items: List[T], min_count: int = 10):
        self.counts = Counter(items)
        relevant_items = [key for key, c in self.counts.items() if c >= min_count]
        self.item_to_index = {key: i + 1 for i, key in enumerate(relevant_items)}
        self.names: List[Union[T, str]] = []
        self.names.append("<UNK>")
        self.names.extend(relevant_items)

    def __len__(self) -> int:
        return len(self.item_to_index) + 1

    def transform_sparse(self, items: List[Any]) -> sps.csr_matrix:
        """Transform sparse"""
        cols = []
        for v in items:
            cols.append(self.item_to_index.get(v, 0))
        rows = np.arange(len(items), dtype=np.int32)
        cols = np.asarray(cols, dtype=np.int32)
        data = np.ones(rows.shape[0], dtype=np.float64)
        result = sps.csr_matrix((data, (rows, cols)), shape=(rows.shape[0], len(self)))

        return result
